# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2016, 2017, 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import logging
import random
import string

from twisted.internet import task
from twisted.internet.defer import Deferred

from ..common import SendRequest

log = logging.getLogger(__name__)

__all__ = [
    'async_delay',
    'first',
    'make_send_requests',
    'random_string',
]


def first(deferreds):
    """Get the first result. Cancel the rest.

    :param deferreds:
        A sequence of `twisted.internet.defer.Deferred` instances.

        Passing a deferred to *first* transfers ownership: the caller must not
        add callbacks or cancel it. *first* cancels all other deferreds as soon
        as one fires.

        This sequence must not be mutated by the caller. *first* does not
        defensively copy.

    :returns: `Deferred` that fires with the result of the first deferred to
        fire or fail. Canceling this deferred cancels all of the deferreds.
    """
    def cancel_all(self):
        for d in deferreds:
            d.cancel()

    result_d = Deferred(cancel_all)

    def one_result(result, source):
        if result_d.called:
            return
        result_d.callback(result)
        for d in deferreds:
            if d is not source:
                d.cancel()

    for d in deferreds:
        d.addBoth(one_result, d)
    return result_d


def random_string(length):
    # Random.choice can be very slow for large amounts of data, so 'cheat'
    if length <= 50:
        s = "".join(random.choice(string.ascii_letters) for _i in range(length))
    else:
        r = random_string(50)
        s = "".join(r for i in range(length // 50))
        if length % 50:
            s += r[0:(length % 50)]
    assert len(s) == length
    return s


def async_delay(timeout=0.01, clock=None):
    if clock is None:
        from twisted.internet import reactor as clock

    return task.deferLater(clock, timeout, lambda: None)


def make_send_requests(msgs, topic=None, key=None):
    return [SendRequest(topic, key, msgs, None)]



# -*- coding: utf-8 -*-
# Copyright (C) 2015 Cyan, Inc.
# Copyright 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
from unittest import skipUnless

from twisted.internet.defer import inlineCallbacks
from twisted.trial import unittest

from afkak import (
    CODEC_GZIP, CODEC_SNAPPY, HashedPartitioner, Producer,
    RoundRobinPartitioner, create_message, create_message_set,
)
from afkak.codec import has_snappy
from afkak.common import (
    PRODUCER_ACK_ALL_REPLICAS, PRODUCER_ACK_LOCAL_WRITE,
    PRODUCER_ACK_NOT_REQUIRED, FetchRequest, ProduceRequest, SendRequest,
)
from afkak.test.testutil import async_delay, make_send_requests, random_string

from .intutil import IntegrationMixin, kafka_versions

log = logging.getLogger(__name__)


class TestAfkakProducerIntegration(IntegrationMixin, unittest.TestCase):
    topic = 'produce_topic'
    harness_kw = dict(replicas=1, partitions=2)

    ###########################################################################
    #   client production Tests  - Server setup is 1 replica, 2 partitions    #
    ###########################################################################

    @kafka_versions("all")
    @inlineCallbacks
    def test_produce_many_simple(self):
        start_offset = yield self.current_offset(self.topic, 0)

        yield self.assert_produce_request(
            [create_message(b"Test message %d" % i) for i in range(100)],
            start_offset,
            100,
        )

        yield self.assert_produce_request(
            [create_message(b"Test message %d" % i) for i in range(100)],
            start_offset + 100,
            100,
        )

    @kafka_versions("all")
    @inlineCallbacks
    def test_produce_100_keyed(self):
        start_offset = yield self.current_offset(self.topic, 0)

        yield self.assert_produce_request(
            [create_message(b"Test message %d" % i,
                            key=b"Key:%d" % i) for i in range(100)],
            start_offset,
            100,
        )
        msgs = [b"Test message %d" % i for i in range(100)]
        keys = [b"Key:%d" % i for i in range(100)]
        yield self.assert_fetch_offset(0, start_offset, msgs,
                                       expected_keys=keys,
                                       fetch_size=10240)

    @kafka_versions("all")
    @inlineCallbacks
    def test_produce_100_keyed_gzipped(self):
        """test_produce_100_keyed_gzipped

        Test that gzipping the batch doesn't affect the partition to which
        the messages are assigned, or the order of the messages in the topics

        """
        start_offset = yield self.current_offset(self.topic, 0)

        msg_set = create_message_set(
            [SendRequest(self.topic, b"Key:%d" % i, [b"Test msg %d" % i], None)
             for i in range(100)], CODEC_GZIP,
        )
        yield self.assert_produce_request(
            msg_set,
            start_offset,
            100,
        )
        msgs = [b"Test msg %d" % i for i in range(100)]
        keys = [b"Key:%d" % i for i in range(100)]
        yield self.assert_fetch_offset(0, start_offset, msgs,
                                       expected_keys=keys,
                                       fetch_size=10240)

    @kafka_versions("all")
    @inlineCallbacks
    def test_produce_10k_simple(self):
        start_offset = yield self.current_offset(self.topic, 0)

        yield self.assert_produce_request(
            [create_message(b"Test message %d" % i) for i in range(10000)],
            start_offset,
            10000,
        )

    @kafka_versions("all")
    @inlineCallbacks
    def test_produce_many_gzip(self):
        start_offset = yield self.current_offset(self.topic, 0)

        message1 = create_message_set(
            make_send_requests(
                [b"Gzipped 1 %d" % i for i in range(100)]), CODEC_GZIP)[0]
        message2 = create_message_set(
            make_send_requests(
                [b"Gzipped 2 %d" % i for i in range(100)]), CODEC_GZIP)[0]

        yield self.assert_produce_request(
            [message1, message2],
            start_offset,
            200,
        )

    @skipUnless(has_snappy(), "Snappy not available")
    @kafka_versions("all")
    @inlineCallbacks
    def test_produce_many_snappy(self):
        start_offset = yield self.current_offset(self.topic, 0)

        message1 = create_message_set(
            make_send_requests(
                [b"Snappy 1 %d" % i for i in range(100)]), CODEC_SNAPPY)[0]
        message2 = create_message_set(
            make_send_requests(
                [b"Snappy 2 %d" % i for i in range(100)]), CODEC_SNAPPY)[0]
        yield self.assert_produce_request(
            [message1, message2],
            start_offset,
            200,
        )

    @kafka_versions("all")
    @inlineCallbacks
    def test_produce_mixed(self):
        start_offset = yield self.current_offset(self.topic, 0)

        msg_count = 1+100
        messages = [
            create_message(b"Just a plain message"),
            create_message_set(make_send_requests([b"Gzipped %d" % i for i in range(100)]), CODEC_GZIP)[0],
        ]

        # Can't produce snappy messages without snappy...
        if has_snappy():
            msg_count += 100
            messages.append(
                create_message_set(
                    make_send_requests([b"Snappy %d" % i for i in range(100)]),
                    CODEC_SNAPPY)[0])

        yield self.assert_produce_request(messages, start_offset, msg_count)
        ptMsgs = [b'Just a plain message']
        ptMsgs.extend([b"Gzipped %d" % i for i in range(100)])
        if has_snappy():
            ptMsgs.extend([b"Snappy %d" % i for i in range(100)])
        yield self.assert_fetch_offset(0, start_offset, ptMsgs,
                                       fetch_size=10240)

    @kafka_versions("all")
    @inlineCallbacks
    def test_produce_10k_gzipped(self):
        start_offset = yield self.current_offset(self.topic, 0)

        msgs = create_message_set(
            make_send_requests(
                [b"Gzipped batch 1, message %d" % i for i in range(5000)]),
            CODEC_GZIP)
        yield self.assert_produce_request(msgs, start_offset, 5000)
        msgs = create_message_set(
            make_send_requests(
                [b"Gzipped batch 2, message %d" % i for i in range(5000)]),
            CODEC_GZIP)
        yield self.assert_produce_request(msgs, start_offset+5000, 5000)

    ###################################################################
    #   Producer Tests  - Server setup is 1 replica, 2 partitions     #
    ###################################################################

    @kafka_versions("all")
    @inlineCallbacks
    def test_producer_send_messages(self):
        start_offset0 = yield self.current_offset(self.topic, 0)
        start_offset1 = yield self.current_offset(self.topic, 1)
        producer = Producer(self.client)

        # Goes to first partition
        resp = yield producer.send_messages(
            self.topic, msgs=[self.msg("one"), self.msg("two")])
        self.assert_produce_response(resp, start_offset0)

        # Goes to the 2nd partition
        resp = yield producer.send_messages(self.topic,
                                            msgs=[self.msg("three")])
        self.assert_produce_response(resp, start_offset1)

        # fetch the messages back and make sure they are as expected
        yield self.assert_fetch_offset(
            0, start_offset0, [self.msg("one"), self.msg("two")])
        yield self.assert_fetch_offset(
            1, start_offset1, [self.msg("three")])
        # Goes back to the first partition because there's only two partitions
        resp = yield producer.send_messages(
            self.topic, msgs=[self.msg("four"), self.msg("five")])
        self.assert_produce_response(resp, start_offset0+2)
        yield self.assert_fetch_offset(
            0, start_offset0, [self.msg("one"),
                               self.msg("two"),
                               self.msg("four"),
                               self.msg("five")])

        yield producer.stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_producer_round_robin_partitioner(self):
        start_offset0 = yield self.current_offset(self.topic, 0)
        start_offset1 = yield self.current_offset(self.topic, 1)

        producer = Producer(self.client,
                            partitioner_class=RoundRobinPartitioner)
        resp1 = yield producer.send_messages(
            self.topic, b"key1", [self.msg("one")])
        resp2 = yield producer.send_messages(
            self.topic, b"key2", [self.msg("two")])
        resp3 = yield producer.send_messages(
            self.topic, b"key3", [self.msg("three")])
        resp4 = yield producer.send_messages(
            self.topic, b"key4", [self.msg("four")])

        self.assert_produce_response(resp1, start_offset0+0)
        self.assert_produce_response(resp2, start_offset1+0)
        self.assert_produce_response(resp3, start_offset0+1)
        self.assert_produce_response(resp4, start_offset1+1)

        yield self.assert_fetch_offset(
            0, start_offset0, [self.msg("one"), self.msg("three")])
        yield self.assert_fetch_offset(
            1, start_offset1, [self.msg("two"), self.msg("four")])

        yield producer.stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_producer_round_robin_partitioner_random_start(self):
        try:
            RoundRobinPartitioner.set_random_start(True)
            producer = Producer(self.client,
                                partitioner_class=RoundRobinPartitioner)
            # Two partitions, so 1st and 3rd reqs should go to same part, but
            # 2nd should go to different. Other than that, without statistical
            # test, we can't say much... Partitioner tests should ensure that
            # we really aren't always starting on a non-random partition...
            resp1 = yield producer.send_messages(
                self.topic, msgs=[self.msg("one"), self.msg("two")])
            resp2 = yield producer.send_messages(
                self.topic, msgs=[self.msg("three")])
            resp3 = yield producer.send_messages(
                self.topic, msgs=[self.msg("four"), self.msg("five")])

            self.assertEqual(resp1.partition, resp3.partition)
            self.assertNotEqual(resp1.partition, resp2.partition)

            yield producer.stop()
        finally:
            RoundRobinPartitioner.set_random_start(False)

    @kafka_versions("all")
    @inlineCallbacks
    def test_producer_hashed_partitioner(self):
        start_offset0 = yield self.current_offset(self.topic, 0)
        start_offset1 = yield self.current_offset(self.topic, 1)

        producer = Producer(self.client,
                            partitioner_class=HashedPartitioner)

        resp1 = yield producer.send_messages(
            self.topic, b'1', [self.msg("one")])
        resp2 = yield producer.send_messages(
            self.topic, b'2', [self.msg("two")])
        resp3 = yield producer.send_messages(
            self.topic, b'1', [self.msg("three")])
        resp4 = yield producer.send_messages(
            self.topic, b'1', [self.msg("four")])
        resp5 = yield producer.send_messages(
            self.topic, b'2', [self.msg("five")])

        self.assert_produce_response(resp2, start_offset0+0)
        self.assert_produce_response(resp5, start_offset0+1)

        self.assert_produce_response(resp1, start_offset1+0)
        self.assert_produce_response(resp3, start_offset1+1)
        self.assert_produce_response(resp4, start_offset1+2)

        yield self.assert_fetch_offset(
            0, start_offset0, [
                self.msg("two"),
                self.msg("five"),
            ],
        )
        yield self.assert_fetch_offset(
            1, start_offset1, [
                self.msg("one"),
                self.msg("three"),
                self.msg("four"),
            ],
        )

        yield producer.stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_producer_batched_gzipped_hashed_partitioner(self):
        start_offset0 = yield self.current_offset(self.topic, 0)
        start_offset1 = yield self.current_offset(self.topic, 1)
        offsets = (start_offset0, start_offset1)

        requests = []
        msgs_by_partition = ([], [])
        keys_by_partition = ([], [])

        partitioner = HashedPartitioner(self.topic, [0, 1])
        producer = Producer(
            self.client, codec=CODEC_GZIP, batch_send=True, batch_every_n=100,
            batch_every_t=None, partitioner_class=HashedPartitioner)

        # Send ten groups of messages, each with a different key
        for i in range(10):
            msg_group = []
            key = 'Key: {}'.format(i).encode()
            part = partitioner.partition(key, [0, 1])
            for j in range(10):
                msg = self.msg('Group:{} Msg:{}'.format(i, j))
                msg_group.append(msg)
                msgs_by_partition[part].append(msg)
                keys_by_partition[part].append(key)
            request = producer.send_messages(
                self.topic, key=key, msgs=msg_group)
            requests.append(request)
            yield async_delay(.5)  # Make the NoResult test have teeth...
            if i < 9:
                # This is to ensure we really are batching all the requests
                self.assertNoResult(request)

        # Now ensure we can retrieve the right messages from each partition
        for part in [0, 1]:
            yield self.assert_fetch_offset(
                part, offsets[part], msgs_by_partition[part],
                keys_by_partition[part], fetch_size=20480)

        yield producer.stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_acks_none(self):
        start_offset0 = yield self.current_offset(self.topic, 0)
        yield self.current_offset(self.topic, 1)

        producer = Producer(
            self.client, req_acks=PRODUCER_ACK_NOT_REQUIRED)
        resp = yield producer.send_messages(self.topic, msgs=[self.msg("one")])
        self.assertEqual(resp, None)

        yield self.assert_fetch_offset(0, start_offset0, [self.msg("one")])
        yield producer.stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_acks_local_write(self):
        start_offset0 = yield self.current_offset(self.topic, 0)
        yield self.current_offset(self.topic, 1)

        producer = Producer(
            self.client, req_acks=PRODUCER_ACK_LOCAL_WRITE)
        resp = yield producer.send_messages(self.topic, msgs=[self.msg("one")])

        self.assert_produce_response(resp, start_offset0)
        yield self.assert_fetch_offset(0, start_offset0, [self.msg("one")])

        yield producer.stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_acks_all_replicas(self):
        start_offset0 = yield self.current_offset(self.topic, 0)
        start_offset1 = yield self.current_offset(self.topic, 1)

        producer = Producer(
            self.client,
            req_acks=PRODUCER_ACK_ALL_REPLICAS)

        resp = yield producer.send_messages(self.topic, msgs=[self.msg("one")])
        self.assert_produce_response(resp, start_offset0)
        yield self.assert_fetch_offset(0, start_offset0, [self.msg("one")])
        yield self.assert_fetch_offset(1, start_offset1, [])

        yield producer.stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_producer_batched_by_messages(self):
        start_offset0 = yield self.current_offset(self.topic, 0)
        start_offset1 = yield self.current_offset(self.topic, 1)

        producer = Producer(self.client, batch_send=True, batch_every_n=10,
                            batch_every_b=0, batch_every_t=0)
        # Send 4 messages and do a fetch. Fetch should timeout, and send
        # deferred shouldn't have a result yet...
        send1D = producer.send_messages(
            self.topic, msgs=[self.msg("one"), self.msg("two"),
                              self.msg("three"), self.msg("four")])
        # by default the assert_fetch_offset() waits for 0.5 secs on the server
        # side before returning no result. So, these two calls should take 1sec
        yield self.assert_fetch_offset(0, start_offset0, [])
        yield self.assert_fetch_offset(1, start_offset1, [])
        # Messages shouldn't have sent out yet, so we shouldn't have
        # response from server yet on having received/responded to the request
        self.assertNoResult(send1D)
        # Sending 3 more messages won't trigger the send either (7 total)
        send2D = producer.send_messages(
            self.topic, msgs=[self.msg("five"), self.msg("six"),
                              self.msg("seven")])
        # make sure no messages on server...
        yield self.assert_fetch_offset(0, start_offset0, [])
        yield self.assert_fetch_offset(1, start_offset1, [])
        # Still no result on send
        self.assertNoResult(send2D)
        # send final batch which will be on partition 0 again...
        send3D = producer.send_messages(
            self.topic, msgs=[self.msg("eight"), self.msg("nine"),
                              self.msg("ten"), self.msg("eleven")])
        # Now do a fetch, again waiting for up to 0.5 seconds for the response
        # All four messages sent in first batch (to partition 0, given default
        # R-R, start-at-zero partitioner), and 4 in 3rd batch
        yield self.assert_fetch_offset(
            0, start_offset0, [self.msg("one"), self.msg("two"),
                               self.msg("three"), self.msg("four"),
                               self.msg("eight"), self.msg("nine"),
                               self.msg("ten"), self.msg("eleven")],
            fetch_size=2048)
        # Fetch from partition:1 should have all messages in 2nd batch
        yield self.assert_fetch_offset(
            1, start_offset1, [self.msg("five"), self.msg("six"),
                               self.msg("seven")])
        # make sure the deferreds fired with the proper result
        resp1 = self.successResultOf(send1D)
        resp2 = self.successResultOf(send2D)
        resp3 = self.successResultOf(send3D)
        self.assert_produce_response(resp1, start_offset0)
        self.assert_produce_response(resp2, start_offset1)
        self.assert_produce_response(resp3, start_offset0)
        # cleanup
        yield producer.stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_producer_batched_by_bytes(self):
        start_offset0 = yield self.current_offset(self.topic, 0)
        start_offset1 = yield self.current_offset(self.topic, 1)

        producer = Producer(self.client, batch_send=True, batch_every_b=4096,
                            batch_every_n=0, batch_every_t=0)
        # Send 4 messages and do a fetch. Fetch should timeout, and send
        # deferred shouldn't have a result yet...
        send1D = producer.send_messages(
            self.topic, msgs=[self.msg("one"), self.msg("two"),
                              self.msg("three"), self.msg("four")])
        # by default the assert_fetch_offset() waits for 0.5 secs on the server
        # side before returning no result. So, these two calls should take 1sec
        yield self.assert_fetch_offset(0, start_offset0, [])
        yield self.assert_fetch_offset(1, start_offset1, [])
        # Messages shouldn't have sent out yet, so we shouldn't have
        # response from server yet on having received/responded to the request
        self.assertNoResult(send1D)
        # Sending 3 more messages should trigger the send, but we don't yield
        # here, so we shouldn't have a response immediately after, and so
        # send2D should still have no result.
        send2D = producer.send_messages(
            self.topic, msgs=[self.msg("five"), self.msg("six"),
                              self.msg("seven")])
        # make sure no messages on server...
        yield self.assert_fetch_offset(0, start_offset0, [])
        yield self.assert_fetch_offset(1, start_offset1, [])
        # Still no result on send
        self.assertNoResult(send2D)
        # send 3rd batch which will be on partition 0 again...
        send3D = producer.send_messages(
            self.topic, msgs=[self.msg("eight"), self.msg("nine"),
                              self.msg("ten"), self.msg("eleven")])
        # make sure no messages on server...
        yield self.assert_fetch_offset(0, start_offset0, [])
        yield self.assert_fetch_offset(1, start_offset1, [])
        # Still no result on send
        self.assertNoResult(send3D)
        # Finally, send a big message to trigger send
        send4D = producer.send_messages(
            self.topic, msgs=[self.msg("1234" * 1024)])
        # Now do a fetch, again waiting for up to 0.5 seconds for the response
        # All four messages sent in first batch (to partition 0, given default
        # R-R, start-at-zero partitioner), and 4 in 3rd batch
        yield self.assert_fetch_offset(
            0, start_offset0, [self.msg("one"), self.msg("two"),
                               self.msg("three"), self.msg("four"),
                               self.msg("eight"), self.msg("nine"),
                               self.msg("ten"), self.msg("eleven")],
            fetch_size=2048)
        # Fetch from partition:1 should have all messages in 2nd batch, as
        # send_messages() treats calls as groups and all/none are sent...
        yield self.assert_fetch_offset(
            1, start_offset1, [self.msg("five"), self.msg("six"),
                               self.msg("seven"), self.msg("1234" * 1024)],
            fetch_size=5*1024)
        # make sure the deferreds fired with the proper result
        resp1 = self.successResultOf(send1D)
        resp2 = self.successResultOf(send2D)
        resp3 = self.successResultOf(send3D)
        resp4 = self.successResultOf(send4D)
        self.assert_produce_response(resp1, start_offset0)
        self.assert_produce_response(resp2, start_offset1)
        self.assert_produce_response(resp3, start_offset0)
        self.assert_produce_response(resp4, start_offset1)
        # cleanup
        yield producer.stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_producer_batched_by_time(self):
        start_offset0 = yield self.current_offset(self.topic, 0)
        start_offset1 = yield self.current_offset(self.topic, 1)

        # This needs to be big enough that the operations between starting the
        # producer and the sleep take less time than this... I made
        # it large enough that the test would still pass even with my Macbook's
        # cores all pegged by a load generator.
        batchtime = 5

        try:
            producer = Producer(self.client, batch_send=True,
                                batch_every_n=0, batch_every_t=batchtime)
            startTime = time.time()
            # Send 4 messages and do a fetch
            send1D = producer.send_messages(
                self.topic, msgs=[self.msg("one"), self.msg("two"),
                                  self.msg("three"), self.msg("four")])
            # set assert_fetch_offset() to wait for 0.1 secs on the server-side
            # before returning no result. So, these calls should take 0.2sec
            yield self.assert_fetch_offset(0, start_offset0, [], max_wait=0.1)
            yield self.assert_fetch_offset(1, start_offset1, [], max_wait=0.1)
            # Messages shouldn't have sent out yet, so we shouldn't have
            # response from server yet on having received/responded to request
            self.assertNoResult(send1D)
            # Sending 3 more messages should NOT trigger the send, as less than
            # 1 sec. elapsed by here, so send2D should still have no result.
            send2D = producer.send_messages(
                self.topic, msgs=[self.msg("five"), self.msg("six"),
                                  self.msg("seven")])
            # still no messages...
            yield self.assert_fetch_offset(0, start_offset0, [], max_wait=0.1)
            yield self.assert_fetch_offset(1, start_offset1, [], max_wait=0.1)
            # Still no result on send, and send should NOT have gone out.
            self.assertNoResult(send2D)
            # Wait the timeout out. It'd be nicer to be able to just 'advance'
            # the reactor, but since we need the network so...
            yield async_delay(batchtime - (time.time() - startTime) + 0.05, clock=self.reactor)
            # We need to yield to the reactor to have it process the response
            # from the broker. Both send1D and send2D should then have results.
            resp1 = yield send1D
            resp2 = yield send2D
            # ensure the 2 batches went into the proper partitions...
            self.assert_produce_response(resp1, start_offset0)
            self.assert_produce_response(resp2, start_offset1)
            # Should be able to get messages now
            yield self.assert_fetch_offset(
                0, start_offset0, [self.msg("one"), self.msg("two"),
                                   self.msg("three"), self.msg("four")])
            yield self.assert_fetch_offset(
                1, start_offset1, [self.msg("five"), self.msg("six"),
                                   self.msg("seven")])
        finally:
            yield producer.stop()

    @kafka_versions("all")
    @inlineCallbacks
    def test_write_nonextant_topic(self):
        """
        Test we can write to a non-extant topic (which will be auto-created)
        simply by calling producer.send_messages with a long enough timeout.
        """
        test_topics = ["{}-{}-{}".format(
            self.id().split('.')[-1], i, random_string(10)) for i in range(10)]

        producer = Producer(self.client, req_acks=PRODUCER_ACK_LOCAL_WRITE)

        for topic in test_topics:
            resp = yield producer.send_messages(topic, msgs=[self.msg(topic)])
            # Make sure the send went ok
            self.assert_produce_response(resp, 0)
            # Make sure we can get the message back
            yield self.assert_fetch_offset(
                0, 0, [self.msg(topic)], topic=topic)

        yield producer.stop()

    # ###########  Utility Functions  ############

    @inlineCallbacks
    def assert_produce_request(self, messages, initial_offset, message_ct):
        produce = ProduceRequest(self.topic, 0, messages=messages)

        # There should only be one response message from the server.
        # This will throw an exception if there's more than one.
        resp, = yield self.client.send_produce_request([produce])
        self.assert_produce_response(resp, initial_offset)

        resp2 = yield self.current_offset(self.topic, 0)

        self.assertEqual(resp2, initial_offset + message_ct)

    def assert_produce_response(self, resp, initial_offset):
        self.assertEqual(resp.error, 0)
        self.assertEqual(resp.offset, initial_offset)

    @inlineCallbacks
    def assert_fetch_offset(self, partition, start_offset,
                            expected_messages, expected_keys=(),
                            max_wait=0.5, fetch_size=1024, topic=None):
        # There should only be one response message from the server.
        # This will throw an exception if there's more than one.
        if topic is None:
            topic = self.topic
        resp, = yield self.client.send_fetch_request(
            [FetchRequest(topic, partition, start_offset, fetch_size)],
            max_wait_time=int(max_wait * 1000))

        self.assertEqual(resp.error, 0)
        self.assertEqual(resp.partition, partition)
        # Convert generator to list
        resp_messages = list(resp.messages)
        messages = [x.message.value for x in resp_messages]
        self.assertEqual(messages, expected_messages)
        if expected_keys:
            keys = [x.message.key for x in resp_messages]
            self.assertEqual(keys, expected_keys)
        self.assertEqual(
            resp.highwaterMark, start_offset+len(expected_messages))



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2017, 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

from mock import ANY, Mock, call, patch
from twisted.internet.defer import CancelledError, Deferred, fail, succeed
from twisted.python.failure import Failure
from twisted.test.proto_helpers import MemoryReactorClock
from twisted.trial import unittest

from .. import consumer as kconsumer  # for patching
from ..common import (
    KAFKA_SUCCESS, OFFSET_COMMITTED, OFFSET_EARLIEST, OFFSET_LATEST,
    TIMESTAMP_INVALID, ConsumerFetchSizeTooSmall, FetchRequest, FetchResponse,
    InvalidConsumerGroupError, KafkaUnavailableError, Message,
    OffsetCommitRequest, OffsetCommitResponse, OffsetFetchRequest,
    OffsetFetchResponse, OffsetOutOfRangeError, OffsetRequest, OffsetResponse,
    OperationInProgress, RestartError, RestopError, SourcedMessage,
    UnknownError,
)
from ..consumer import FETCH_BUFFER_SIZE_BYTES, Consumer
from ..kafkacodec import KafkaCodec, create_message
from .logtools import capture_logging

log = logging.getLogger(__name__)


class TestAfkakConsumer(unittest.SynchronousTestCase):
    maxDiff = None

    def test_consumer_non_integer_partitions(self):
        with self.assertRaises(ValueError):
            Consumer(Mock(), 'topic', '0', Mock())

    def test_consumer_non_integer_commit_every_n(self):
        with self.assertRaises(ValueError):
            Consumer(
                Mock(reactor=MemoryReactorClock()), 'topic', 0, Mock(),
                consumer_group='test_consumer_non_integer_commit_every_n',
                auto_commit_every_n=3.5,
            )

    def test_consumer_negative_commit_every_n(self):
        with self.assertRaises(ValueError):
            Consumer(
                Mock(reactor=MemoryReactorClock()), 'topic', 0, Mock(),
                consumer_group='test_consumer_negative_commit_every_n',
                auto_commit_every_n=-300,
            )

    def test_consumer_non_integer_commit_every_ms(self):
        with self.assertRaises(ValueError):
            Consumer(
                Mock(), 'topic', 0, Mock(),
                consumer_group='test_consumer_non_integer_commit_every_ms',
                auto_commit_every_ms=3.5,
            )

    def test_consumer_negative_commit_every_ms(self):
        with self.assertRaises(ValueError):
            Consumer(
                Mock(), 'topic', 0, Mock(),
                consumer_group='test_consumer_negative_commit_every_ms',
                auto_commit_every_ms=-20,
            )

    def test_consumer_non_integer_retry_max_attempts(self):
        with self.assertRaises(ValueError):
            Consumer(
                Mock(), 'topic', 0, Mock(),
                consumer_group='test_consumer_non_integer_retry_max_attempts',
                request_retry_max_attempts=20.3,
            )

    def test_consumer_negative_retry_max_attempts(self):
        with self.assertRaises(ValueError):
            Consumer(
                Mock(), 'topic', 0, Mock(),
                consumer_group='test_consumer_negative_retry_max_attempts',
                request_retry_max_attempts=-20,
            )

    def test_consumer_non_str_auto_offset_reset(self):
        with self.assertRaises(ValueError):
            Consumer(
                Mock(), 'topic', 0, Mock(),
                consumer_group='test_consumer_non_str_auto_offset_reset',
                auto_offset_reset=111,
            )

    def test_consumer_str_not_expected(self):
        with self.assertRaises(ValueError):
            Consumer(
                Mock(), 'topic', 0, Mock(),
                consumer_group='test_consumer_str_not_expected',
                auto_offset_reset="not_expected",
            )

    def test_consumer_init(self):
        client = Mock(reactor=MemoryReactorClock())
        partition = 9
        processor = Mock()
        consumer_group = 'My Consumer Group'
        consumer_metadata = b'My Commit Metadata'
        auto_commit_msgs = 24
        auto_commit_time = 60000

        Consumer(
            client, 'tTopic', partition, processor, consumer_group,
            consumer_metadata, auto_commit_msgs, auto_commit_time,
            4096, 1000, 256 * 1024, 8 * 1024 * 1024, 1.0, 30,
        )

    def test_consumer_buffer_size_err(self):
        with self.assertRaises(ValueError):
            Consumer(None, 'Grues', 99, Mock(), buffer_size=8192,
                     max_buffer_size=4096)

    def test_consumer_auto_commit_parms_err(self):
        with self.assertRaises(ValueError):
            Consumer(None, 'Agnot', 500, Mock(), auto_commit_every_ms=8192)

    def test_consumer_repr(self):
        mockClient = Mock(reactor=MemoryReactorClock())
        consumer = Consumer(mockClient, 'Grues', 99, lambda c, m: None)
        self.assertEqual('<Consumer Grues/99 initialized>', repr(consumer))

    def test_consumer_start_offset(self):
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        consumer = Consumer(mockclient, u'offset22Topic', 18, Mock())
        d = consumer.start(22)
        request = FetchRequest('offset22Topic', 18, 22, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        self.assertIsNone(consumer.stop())
        self.assertIsNone(self.successResultOf(d))

    def test_consumer_start_earliest(self):
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        consumer = Consumer(mockclient, 'earliestTopic', 9, Mock())
        d = consumer.start(OFFSET_EARLIEST)
        request = OffsetRequest(u'earliestTopic', 9, OFFSET_EARLIEST, 1)
        mockclient.send_offset_request.assert_called_once_with([request])
        self.assertIsNone(consumer.stop())
        self.assertIsNone(self.successResultOf(d))

    def test_consumer_start_latest(self):
        offset = 2346  # arbitrary
        topic = 'latestTopic'
        part = 10
        reqs_ds = [Deferred(), Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_offset_request.return_value = reqs_ds[0]
        mockclient.send_offset_fetch_request.return_value = reqs_ds[1]
        consumer = Consumer(mockclient, topic, part, Mock())
        d = consumer.start(OFFSET_LATEST)
        # Make sure request was made
        request = OffsetRequest(topic, part, OFFSET_LATEST, 1)
        mockclient.send_offset_request.assert_called_once_with([request])
        # Deliver the responses
        responses = [OffsetResponse(topic, part, KAFKA_SUCCESS, [offset])]
        reqs_ds[0].callback(responses)
        self.assertEqual(offset, consumer._fetch_offset)
        # Check that the message fetch was started
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # Stop the consumer to cleanup any outstanding operations
        self.assertIsNone(consumer.stop())
        self.assertIsNone(self.successResultOf(d))

    def test_consumer_start_committed(self):
        offset = 2996  # arbitrary, offset we're committing
        fetch_offset = offset + 1  # fetch at next offset after committed
        topic = u'committedTopic'
        part = 23
        reqs_ds = [Deferred(), Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_offset_fetch_request.return_value = reqs_ds[0]
        mockclient.send_fetch_request.return_value = reqs_ds[1]
        consumer = Consumer(mockclient, topic, part, Mock(),
                            consumer_group=u"myGroup")
        d = consumer.start(OFFSET_COMMITTED)
        # Make sure request was made
        request = OffsetFetchRequest(topic, part)
        mockclient.send_offset_fetch_request.assert_called_once_with(
            'myGroup', [request])
        # Deliver the response
        responses = [OffsetFetchResponse(topic, part, offset, b"METADATA",
                                         KAFKA_SUCCESS)]
        reqs_ds[0].callback(responses)
        self.assertEqual(fetch_offset, consumer._fetch_offset)
        # Check that the message fetch was started
        request = FetchRequest(topic, part, fetch_offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # Stop the consumer to cleanup any outstanding operations
        self.assertIsNone(consumer.stop())
        self.assertIsNone(self.successResultOf(d))

    def test_consumer_start_committed_bad_group(self):
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        consumer = Consumer(mockclient, 'committedTopic', 11, Mock())
        d = consumer.start(OFFSET_COMMITTED)
        self.assertFalse(mockclient.called)
        self.failureResultOf(d, InvalidConsumerGroupError)

    def test_consumer_commit_bad_group(self):
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        consumer = Consumer(mockclient, 'committedTopic', 11, Mock())
        d = consumer.commit()
        self.assertFalse(mockclient.called)
        self.failureResultOf(d, InvalidConsumerGroupError)

    def test_consumer_commit_no_progress(self):
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        consumer = Consumer(mockclient, 'committedTopic', 11, Mock(), 'cGroup')
        d = consumer.commit()
        self.assertFalse(mockclient.called)
        self.assertEqual(self.successResultOf(d), None)

    def test_consumer_commit_with_progress(self):
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        return_value = Deferred()
        mockclient.send_offset_commit_request.return_value = return_value
        the_group = 'Band on the Run'
        the_topic = 'test_consumer_commit_with_progress_topic'
        the_part = 1134
        the_offset = 4269
        the_request = OffsetCommitRequest(
            the_topic, the_part, the_offset, TIMESTAMP_INVALID, None)
        # Create a consumer and muck with the state a bit...
        consumer = Consumer(mockclient, the_topic, the_part, Mock(), the_group)
        consumer._last_processed_offset = the_offset  # Fake processed msgs
        consumer._commit_looper = Mock()  # Mock a looping call to test reset
        d = consumer.commit()
        mockclient.send_offset_commit_request.assert_called_once_with(
            the_group, [the_request], consumer_id='', group_generation_id=-1)
        consumer._commit_looper.reset.assert_called_once_with()
        self.assertNoResult(d)

    def test_consumer_commit_during_commit(self):
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        return_value = Deferred()
        mockclient.send_offset_commit_request.return_value = return_value
        the_group = 'The Cure'
        the_topic = 'test_consumer_commit_during_commit_topic'
        the_part = 1
        the_offset = 28616
        the_request = OffsetCommitRequest(
            the_topic, the_part, the_offset, TIMESTAMP_INVALID, None)
        # Create a consumer and muck with the state a bit...
        consumer = Consumer(mockclient, the_topic, the_part, Mock(), the_group)
        consumer._last_processed_offset = the_offset  # Fake processed msgs
        consumer._commit_looper = Mock()  # Mock a looping call to test reset
        d1 = consumer.commit()
        mockclient.send_offset_commit_request.assert_called_once_with(
            the_group, [the_request], consumer_id='', group_generation_id=-1)
        consumer._commit_looper.reset.assert_called_once_with()
        self.assertFalse(d1.called)
        d2 = consumer.commit()
        self.failureResultOf(d2, OperationInProgress)

    def test_consumer_auto_commit_by_msgs(self):
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        client_requests = [Deferred(), Deferred()]
        mockclient.send_fetch_request.return_value = client_requests[0]
        mockclient.send_offset_commit_request.return_value = client_requests[1]
        the_group = u'Horse with no name'
        the_topic = 'test_consumer_auto_commit_by_msgs'
        the_part = 1341
        the_offset = 2694
        the_processor = Mock()
        proc_deferreds = [Deferred(), Deferred(), Deferred(), Deferred(),
                          Deferred(), Deferred()]
        the_processor.side_effect = proc_deferreds
        # Create a consumer and start it at offset 0
        consumer = Consumer(mockclient, the_topic, the_part,
                            the_processor, the_group, auto_commit_every_n=1,
                            auto_commit_every_ms=0)  # No auto_commit by time
        start_d = consumer.start(the_offset)
        # Fire a response to the fetch request
        messages = [create_message(m) for m in [b"msg1", b"hi", b"boo", b"foo", b"fun"]]
        message_set = KafkaCodec._encode_message_set(messages, the_offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(the_topic, the_part, KAFKA_SUCCESS, 486,
                                   message_iter)]
        client_requests[0].callback(responses)
        # Batch of messages delivered, expect the_processor to have been called
        the_processor.assert_called_once_with(
            consumer,
            [SourcedMessage(the_topic, the_part, the_offset, messages[0])])
        # Finish the processing of the processor
        proc_deferreds[0].callback(True)
        # Expect a commit request
        the_request = OffsetCommitRequest(
            the_topic, the_part, the_offset, TIMESTAMP_INVALID, None)
        mockclient.send_offset_commit_request.assert_called_once_with(
            the_group, [the_request], consumer_id='', group_generation_id=-1)
        # 'Send' the commit response
        commit_response = [
            OffsetCommitResponse(the_topic, the_part, KAFKA_SUCCESS),
        ]
        client_requests[1].callback(commit_response)

        # Stop the consumer to cleanup any outstanding operations
        self.assertNoResult(start_d)
        last_processed = consumer.stop()
        self.assertEqual(self.successResultOf(start_d), the_offset)
        self.assertEqual(last_processed, the_offset)

    def test_consumer_commit_retry(self):
        mockclient = Mock(reactor=MemoryReactorClock())
        commit_ds = [fail(KafkaUnavailableError()), Deferred()]
        mockclient.send_offset_commit_request.side_effect = commit_ds
        the_group = 'Sade'
        the_topic = u'test_consumer_commit_retry'
        the_part = 19
        the_offset = 5431
        the_request = OffsetCommitRequest(
            the_topic, the_part, the_offset, TIMESTAMP_INVALID, None)
        # Create a consumer and muck with the state a bit...
        consumer = Consumer(mockclient, the_topic, the_part, Mock(), the_group)
        consumer._last_processed_offset = the_offset  # Fake processed msgs
        d = consumer.commit()
        mockclient.send_offset_commit_request.assert_called_once_with(
            the_group, [the_request], consumer_id='', group_generation_id=-1)
        mockclient.reactor.advance(consumer.retry_max_delay)
        the_call = call(the_group, [the_request], consumer_id='', group_generation_id=-1)
        expected_calls = [the_call, the_call]
        self.assertEqual(mockclient.send_offset_commit_request.mock_calls,
                         expected_calls)
        commit_response = [OffsetCommitResponse(
            the_topic, the_part, KAFKA_SUCCESS)]
        self.assertFalse(d.called)
        commit_ds[1].callback(commit_response)
        self.assertTrue(d.called)

    def test_consumer_auto_commit_fail_errbacks_start_d(self):
        """
        A failure to auto-commit is reported on the Deferred returned by `start()`.
        """
        mockclient = Mock(reactor=MemoryReactorClock())
        the_group = u'The Clash'
        the_topic = 'test_consumer_auto_commit_fail_errbacks_start_d'
        the_part = 20
        the_offset = 989
        the_request = OffsetCommitRequest(the_topic, the_part, the_offset, TIMESTAMP_INVALID, None)
        # Make the commit fail with something that won't be retried
        the_fail = Failure(ValueError(the_topic))
        fetch_d = Deferred()
        mockclient.send_offset_commit_request.return_value = fail(the_fail)
        mockclient.send_fetch_request.return_value = fetch_d
        # Create a consumer and muck with the state a bit...
        consumer = Consumer(mockclient, the_topic, the_part, Mock(), the_group)
        consumer._last_processed_offset = the_offset  # Fake processed msgs
        # Start the consumer
        start_d = consumer.start(0)
        # fake an _commit_looper call
        consumer._auto_commit()
        # Make sure it tried to commit
        mockclient.send_offset_commit_request.assert_called_once_with(
            the_group, [the_request], consumer_id='', group_generation_id=-1)
        # Make sure the start_d was errback'd
        self.assertEqual(self.failureResultOf(start_d, ValueError), the_fail)
        # Clean up
        consumer.stop()

    def test_consumer_commit_retry_to_failure(self):
        """
        The consumer eventually gives up trying to commit offsets when the
        retry count is exhausted.
        """

        def make_fail(*_, **__):
            return fail(KafkaUnavailableError('commit_retry_to_failure'))

        commit_attempts = 12  # gets us two warnings
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_offset_commit_request.side_effect = make_fail

        the_group = 'TearsForFears'
        the_topic = 'test_consumer_commit_retry_to_failure'
        the_part = 1
        the_offset = 4513
        the_request = OffsetCommitRequest(
            the_topic, the_part, the_offset, TIMESTAMP_INVALID, None)

        # Create a consumer and muck with the state a bit...
        # Also setup the retry timing to give expected retries/log calls
        consumer = Consumer(
            mockclient, the_topic, the_part, Mock(), the_group,
            request_retry_init_delay=1.20205,
            request_retry_max_delay=4.0,
            request_retry_max_attempts=commit_attempts)
        consumer._last_processed_offset = the_offset  # Fake processed msgs
        commit_d = consumer.commit()
        with capture_logging(kconsumer.log) as records:
            while not commit_d.called:
                clock.advance(consumer.retry_max_delay)

        # At least one warning was logged.
        self.assertIn("WARNING", [r.levelname for r in records])

        # Make sure we retried the request the proper number of times
        the_call = call(the_group, [the_request], consumer_id='', group_generation_id=-1)
        expected_calls = [the_call] * commit_attempts
        self.assertEqual(mockclient.send_offset_commit_request.mock_calls,
                         expected_calls)

        self.failureResultOf(commit_d, KafkaUnavailableError)

    def test_consumer_start_twice(self):
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        consumer = Consumer(mockclient, 'twice_start', 12, Mock())
        consumer.start(0)
        self.assertRaises(RestartError, consumer.start, 0)

    def test_consumer_stop_during_offset(self):
        topic = 'stop_during_offset'
        part = 101
        reqs_ds = [Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_offset_request.return_value = reqs_ds[0]
        consumer = Consumer(mockclient, topic, part, Mock())
        d = consumer.start(OFFSET_LATEST)
        # Make sure request was made
        request = OffsetRequest(topic, part, OFFSET_LATEST, 1)
        mockclient.send_offset_request.assert_called_once_with([request])
        # Stop the consumer to cleanup any outstanding operations
        self.assertIsNone(consumer.stop())
        self.assertIsNone(self.successResultOf(d))

    def test_consumer_stop_before_fetch_response(self):
        """test_consumer_stop_before_fetch_response

        BPPF-472: Consumer would enter a tight-loop, blocking reactor by
        continuous, recursive adding of `_handle_fetch_response` from
        within it. This was due to the `self._msg_block_d` being
        cancelled, but not cleared.  This test ensures that if
        `_handle_fetch_response` is entered after the consumer has been
        stopped, that this recursive infinite loop does not occur.
        """
        def processor(consumer, messages):
            # Stop the consumer.
            consumer.stop()

        def make_response(s_id):
            # Create a response to the fetch request
            messages = [create_message(u"msg{}".format(n).encode('ascii'))
                        for n in range(s_id, s_id + 4)]
            message_set = KafkaCodec._encode_message_set(messages, offset)
            message_iter = KafkaCodec._decode_message_set_iter(message_set)
            return [FetchResponse(
                topic, part, KAFKA_SUCCESS, 486, message_iter)]

        topic = 'test_consumer_stop_before_fetch_response'
        part = 201
        offset = 44
        req_ds = [Deferred(), Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_fetch_request.side_effect = req_ds
        consumer = Consumer(mockclient, topic, part, processor)
        start_d = consumer.start(offset)
        # Make sure request was made
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # Fire a response to the fetch request
        req_ds[0].callback(make_response(offset))
        self.assertIsNone(self.successResultOf(start_d))
        clock.advance(consumer.retry_max_delay)
        expected_calls = [
            call([request], max_wait_time=consumer.fetch_max_wait_time,
                 min_bytes=consumer.fetch_min_bytes)]  # Unfixed code makes 2nd req.
        self.assertEqual(mockclient.send_fetch_request.mock_calls,
                         expected_calls)
        req_ds[1].callback(make_response(offset + 4))  # Unfixed code hangs here

    def test_consumer_stop_during_fetch_retry(self):
        fetch_ds = [Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_fetch_request.side_effect = fetch_ds
        consumer = Consumer(mockclient, 'committedTopic', 11, "FakeProc",
                            consumer_group="myGroup")
        d = consumer.start(0)
        with patch.object(kconsumer, 'log') as klog:
            f = Failure(UnknownError())
            fetch_ds[0].errback(f)
            klog.debug.assert_called_once_with(
                "%r: Failure fetching messages from kafka: %r",
                consumer, f)
        self.assertIsNone(consumer.stop())
        self.assertIsNone(self.successResultOf(d))

    def test_consumer_offset_fetch_retry_to_failure(self):
        """
        The consumer eventually gives up trying to load offsets.
        """

        def make_fail(*_, **__):
            return fail(KafkaUnavailableError('offset_fetch_retry_to_failure'))

        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        the_processor = Mock()
        the_topic = 'test_consumer_offset_fetch_retry_to_failure_topic'
        the_part = 13
        fetch_offset = OFFSET_EARLIEST
        fetch_attempts = 100
        mockclient.send_offset_request.side_effect = make_fail
        consumer = Consumer(mockclient, the_topic, the_part, the_processor,
                            request_retry_max_attempts=fetch_attempts)
        request = OffsetRequest(the_topic, the_part, fetch_offset, 1)

        d = consumer.start(fetch_offset)
        while not d.called:
            clock.advance(consumer.retry_max_delay)

        self.failureResultOf(d, KafkaUnavailableError)
        offset_call = call([request])
        self.assertEqual(mockclient.send_offset_request.mock_calls,
                         [offset_call] * fetch_attempts)
        self.assertIsNone(consumer.stop())

    def test_consumer_fetch_retry_to_failure(self):

        def make_fail(*_, **__):
            return fail(KafkaUnavailableError('fetch_retry_to_failure'))

        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        the_processor = Mock()
        the_topic = 'test_consumer_fetch_retry_to_failure_topic'
        the_part = 12
        fetch_offset = 0
        fetch_attempts = 100
        mockclient.send_fetch_request.side_effect = make_fail
        consumer = Consumer(mockclient, the_topic, the_part, the_processor,
                            request_retry_max_attempts=fetch_attempts)
        request = FetchRequest(the_topic, the_part, fetch_offset,
                               consumer.buffer_size)

        d = consumer.start(fetch_offset)
        while not d.called:
            clock.advance(consumer.retry_max_delay)

        self.failureResultOf(d, KafkaUnavailableError)
        fetch_call = call(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        self.assertEqual(mockclient.send_fetch_request.mock_calls,
                         [fetch_call] * fetch_attempts)
        self.assertIsNone(consumer.stop())

    def test_consumer_stop_during_initial_proc_call(self):
        # processor's deferred
        proc_d = Deferred(Mock())
        pmock_errback = Mock()
        proc_d.addErrback(pmock_errback)
        proc_called = [False]

        def processor(consumer, msglist):
            proc_called[0] = True
            consumer.stop()
            return proc_d

        topic = u'proc_stop'
        part = 33
        offset = 67

        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        fetch_ds = [Deferred(), Deferred()]
        mockclient.send_fetch_request.side_effect = fetch_ds
        consumer = Consumer(mockclient, topic, part, processor)
        start_d = consumer.start(offset)
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # create & deliver the response
        messages = [
            create_message(b"v9", b"k9"),
            create_message(b"v10", b"k10"),
        ]
        message_set = KafkaCodec._encode_message_set(messages, offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 486,
                                   message_iter)]
        fetch_ds[0].callback(responses)
        # Make sure the processor was called
        self.assertTrue(proc_called[0])
        # Make sure the processor deferred was cancelled
        proc_d._canceller.assert_called_once_with(proc_d)
        # Make sure processor errback was called (canceller didn't callback)
        self.assertTrue(pmock_errback.called)

        # Make sure the start callback was called, and the errback wasn't
        self.assertIsNone(self.successResultOf(start_d))

    def test_consumer_stop_during_commit_retry(self):
        # setup a client which will return a message block in response to fetch
        # and just fail on the commit
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_offset_commit_request.return_value = fail(
            KafkaUnavailableError())
        fetch_d = Deferred()
        mockclient.send_fetch_request.return_value = fetch_d
        the_group = u'Duran Duran'
        the_topic = u'test_consumer_stop_during_commit_retry'
        the_part = 11
        the_offset = 0
        the_highwater = 5
        # Create a consumer with autocommit disabled
        mock_proc = Mock()
        consumer = Consumer(
            mockclient, the_topic, the_part, mock_proc, the_group,
            auto_commit_every_n=0, auto_commit_every_ms=0)
        # Start the consumer at offset zero
        d = consumer.start(the_offset)

        # create & deliver the response
        messages = [create_message(b"aotearoa"), create_message(b"bikini")]
        message_set = KafkaCodec._encode_message_set(messages, the_offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(
            the_topic, the_part, KAFKA_SUCCESS, the_highwater, message_iter)]
        fetch_d.callback(responses)

        mock_proc.assert_called_once_with(
            consumer,
            [SourcedMessage(the_topic, 11, 0, Message(0, 0, None, b'aotearoa')),
             SourcedMessage(the_topic, 11, 1, Message(0, 0, None, b'bikini'))])

        commit_d = consumer.commit()
        self.assertNoResult(commit_d)

        self.assertEqual(1, consumer.stop())
        self.assertEqual(1, self.successResultOf(d))
        # Now the commit_d should have been cancelled, check for the failure
        self.failureResultOf(commit_d, CancelledError)

    def test_consumer_stop_during_commit(self):
        # setup a client which will return a message block in response to fetch
        # and just fail on the commit
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_offset_commit_request.return_value = Deferred()
        mockclient.send_fetch_request.return_value = Deferred()
        the_group = 'U2'
        the_topic = u'test_consumer_stop_during_commit'
        the_part = 11
        the_offset = 0
        # Create a consumer and muck with the state a bit...
        consumer = Consumer(mockclient, the_topic, the_part, Mock(), the_group,
                            auto_commit_every_ms=0)
        start_d = consumer.start(the_offset)
        consumer._last_processed_offset = the_offset  # Fake processed msgs

        # Start a commit, don't fire the deferred, assert there's no result
        commit_d = consumer.commit()
        self.assertNoResult(commit_d)
        self.assertEqual(consumer._commit_ds[0], commit_d)

        # Stop the consumer, assert the start_d fired, and commit_d errbacks
        self.assertEqual(the_offset, consumer.stop())
        self.assertEqual(the_offset, self.successResultOf(start_d))
        self.failureResultOf(commit_d, CancelledError)

    def test_consumer_stop_not_started(self):
        mockclient = Mock(reactor=MemoryReactorClock())
        consumer = Consumer(mockclient, 'stop_no_start', 12, Mock())
        self.assertRaises(RestopError, consumer.stop)

    def test_consumer_processor_error(self):
        reqs_ds = [Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        proc_d = Deferred()

        topic = u'proc_error'
        part = 30
        offset = 38

        mockclient.send_fetch_request.side_effect = reqs_ds
        consumer = Consumer(mockclient, topic, part, lambda *args, **kwargs: proc_d)
        d = consumer.start(offset)
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # create & deliver the response
        messages = [
            create_message(b"v1", b"k1"),
            create_message(b"v2", b"k2"),
        ]
        message_set = KafkaCodec._encode_message_set(messages, offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 99,
                                   message_iter)]
        reqs_ds[0].callback(responses)
        # Make sure the processor was called
        self.assertEqual(proc_d, consumer._processor_d)

        # Errback the processor deferred
        f = Failure(KeyError())  # Pick some random failure mode
        proc_d.errback(f)
        # Ensure the start() deferred was errback'd
        self.assertEqual(self.failureResultOf(d), f)
        self.assertIsNone(consumer.stop())
        self.assertIsNone(consumer.last_processed_offset)

    def test_consumer_error_during_offset(self):
        topic = 'error_during_offset'
        part = 991
        reqs_ds = [Deferred(), Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_offset_request.side_effect = reqs_ds
        consumer = Consumer(mockclient, topic, part, Mock())
        d = consumer.start(OFFSET_LATEST)
        # Make sure request for offset was made
        request = OffsetRequest(topic, part, OFFSET_LATEST, 1)
        mockclient.send_offset_request.assert_called_once_with([request])
        # Errback the first request
        f = Failure(KafkaUnavailableError())  # Perhaps kafka wasn't up yet...
        with patch.object(kconsumer, 'log'):
            reqs_ds[0].errback(f)
        # Advance the clock to trigger the 2nd request
        clock.advance(consumer.retry_delay + 1)  # fire the callLater
        self.assertEqual(2, mockclient.send_offset_request.call_count)

        # Stop the consumer to cleanup any outstanding operations
        self.assertIsNone(consumer.stop())
        self.assertIsNone(self.successResultOf(d))

    def test_consumer_offset_out_of_range_error_with_auto_reset_to_earliest(self):
        topic = 'offset_out_of_range_error'
        part = 911
        offset = 10000
        reqs_d = Deferred()
        mockclient = Mock()
        mockclient.send_fetch_request.return_value = reqs_d

        consumer = Consumer(mockclient, topic, part, Mock(), auto_offset_reset=OFFSET_EARLIEST)
        consumer.start(offset)

        self.assertIsNone(consumer._retry_call)
        fetch_request = FetchRequest(topic=topic, partition=part, offset=offset,
                                     max_bytes=FETCH_BUFFER_SIZE_BYTES)
        consumer.client.send_fetch_request.assert_called_once_with([fetch_request], max_wait_time=100, min_bytes=65536)

        f = Failure(OffsetOutOfRangeError())
        reqs_d.errback(f)

        self.assertEqual(consumer._fetch_offset, OFFSET_EARLIEST)
        self.assertIsNotNone(consumer._retry_call)

        earliest_offset_request = OffsetRequest(topic, part, OFFSET_EARLIEST, 1)

        with patch.object(kconsumer, 'log'):
            consumer._do_fetch()

        consumer.client.send_offset_request.assert_called_once_with([earliest_offset_request])

        consumer.stop()

    def test_consumer_offset_out_of_range_error_with_auto_reset_to_latest(self):
        topic = 'offset_out_of_range_error'
        part = 911
        offset = 10000
        reqs_d = Deferred()
        mockclient = Mock()
        mockclient.send_fetch_request.side_effect = reqs_d
        consumer = Consumer(mockclient, topic, part, Mock(), auto_offset_reset=OFFSET_LATEST)
        consumer.start(offset)

        self.assertIsNone(consumer._retry_call)
        fetch_request = FetchRequest(topic=topic, partition=part, offset=offset,
                                     max_bytes=FETCH_BUFFER_SIZE_BYTES)
        consumer.client.send_fetch_request.assert_called_once_with([fetch_request], max_wait_time=100, min_bytes=65536)

        f = Failure(OffsetOutOfRangeError())
        reqs_d.errback(f)

        self.assertEqual(consumer._fetch_offset, OFFSET_LATEST)
        self.assertIsNotNone(consumer._retry_call)

        latest_offset_request = OffsetRequest(topic, part, OFFSET_LATEST, 1)

        with patch.object(kconsumer, 'log'):
            consumer._do_fetch()

        consumer.client.send_offset_request.assert_called_once_with([latest_offset_request])

        consumer.stop()

    def test_consumer_offset_out_of_range_error_without_reset(self):
        topic = 'offset_out_of_range_error'
        part = 911
        offset = 10000
        fetch_ds = [Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_fetch_request.side_effect = fetch_ds

        consumer = Consumer(mockclient, topic, part, Mock())
        d = consumer.start(offset)

        f = Failure(OffsetOutOfRangeError())
        fetch_ds[0].errback(f)

        self.assertEqual(self.failureResultOf(d), f)

        consumer.stop()

    def test_consumer_errors_during_offset(self):
        attempts = 5
        topic = 'all_errors_during_offset'
        part = 991
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_offset_request.side_effect = Deferred
        # The request we expect...
        request = OffsetRequest(topic, part, OFFSET_EARLIEST, 1)
        # The error we'll return
        f = Failure(KafkaUnavailableError())  # Perhaps kafka wasn't up yet...

        consumer = Consumer(mockclient, topic, part, Mock(),
                            request_retry_max_attempts=attempts)
        d = consumer.start(OFFSET_EARLIEST)
        # Make sure request for offset was made with correct request
        mockclient.send_offset_request.assert_called_once_with([request])
        call_count = 0
        while not d.called:
            # Make sure more requests are made each time the timer expires
            call_count += 1
            self.assertEqual(call_count,
                             mockclient.send_offset_request.call_count)
            # Errback the request
            consumer._request_d.errback(f)
            # Advance the clock to trigger the next request
            clock.advance(consumer.retry_delay + 0.01)

        self.assertEqual(attempts, call_count)
        # Make sure the start() deferred was errbacked with the failure
        self.assertIs(f, self.failureResultOf(d))
        # Stop the consumer to cleanup any outstanding operations
        consumer.stop()

    def test_consumer_fetch_reply_during_processing(self):
        fetch_ds = [Deferred(), Deferred(), Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        proc_ds = [Deferred(), Deferred()]
        deferMaker = Mock()
        deferMaker.side_effect = proc_ds

        topic = 'repl_during_proc'
        part = 42
        offset = 1967

        mockclient.send_fetch_request.side_effect = fetch_ds
        consumer = Consumer(mockclient, topic, part, deferMaker)
        d = consumer.start(offset)
        messages = [create_message(b"v9", b"k9"), create_message(b"v10", b"k10")]

        # Return a message set starting 1 before the requested offest to
        # exercise the message-skipping code in Consumer._handle_fetch_response
        # which deals with the fact that Kafka can return messages with offests
        # less than requested due to messages being compressed as a set, and
        # the whole compressed set being returned together
        message_set = KafkaCodec._encode_message_set(messages, offset - 1)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 486,
                                   message_iter)]
        with patch.object(kconsumer, 'log'):
            fetch_ds[0].callback(responses)
        # Make sure the processor was called
        self.assertEqual(proc_ds[0], consumer._processor_d)
        # Trigger another fetch
        clock.advance(0.01)
        # Make sure the consumer made a 2nd fetch request
        self.assertEqual(fetch_ds[1], consumer._request_d)
        # Make sure the consumer is still waiting on the 1st processor deferred
        self.assertEqual(proc_ds[0], consumer._processor_d)
        # Deliver the 2nd fetch result
        message_set = KafkaCodec._encode_message_set(messages, offset + 1)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 486,
                                   message_iter)]
        fetch_ds[1].callback(responses)
        # Make sure the consumer is STILL waiting on the 1st processor deferred
        self.assertEqual(proc_ds[0], consumer._processor_d)
        # And is STILL waiting on the 2nd fetch reply
        self.assertEqual(fetch_ds[1], consumer._request_d)

        # Deliver the processing result
        proc_ds[0].callback(None)
        # Confirm the consumer is now waiting on the 2nd processor deferred,
        # and isn't waiting on any fetch result
        self.assertEqual(proc_ds[1], consumer._processor_d)
        self.assertEqual(None, consumer._request_d)

        # stop consumer to clean up
        self.assertEqual(offset, consumer.stop())
        self.assertEqual(offset, self.successResultOf(d))
        self.assertEqual(offset, consumer.last_processed_offset)
        self.assertIsNone(consumer.last_committed_offset)

    def test_consumer_fetch_large_message(self):
        topic = 'fetch_large_message'
        part = 676
        offset = 0
        mock_proc = Mock()
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        reqs_ds = [Deferred() for x in range(10)]
        mockclient.send_fetch_request.side_effect = reqs_ds

        consumer = Consumer(mockclient, topic, part, mock_proc)
        msg_size = consumer.buffer_size * 9
        messages = [create_message(b'*' * msg_size)]
        message_set = KafkaCodec._encode_message_set(messages, offset)
        d = consumer.start(offset)
        log.debug("Started Consumer: %r start_d: %r", consumer, d)

        # Ok, we deliver only part of the message_set, up to the size requested
        while not mock_proc.called:
            # Get the buffer size from the last call
            request = mockclient.send_fetch_request.call_args[0][0][0]
            log.debug("Got request: %r msg size: %d", request, msg_size)
            # Create a response only as large as the 'max_bytes' request param
            message_iter = KafkaCodec._decode_message_set_iter(
                message_set[0:request.max_bytes])
            responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 486,
                                       message_iter)]
            log.debug("Calling _request_d: %r callback with: %r",
                      consumer._request_d, responses)
            consumer._request_d.callback(responses)
            # Advance the clock to trigger the next request
            clock.advance(0.1)

        self.assertEqual(0, consumer.stop())
        self.assertEqual(0, self.successResultOf(d))

    def test_consumer_fetch_too_large_message(self):
        topic = 'fetch_too_large_message'
        part = 676
        offset = 0
        mock_proc = Mock()
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        reqs_ds = [Deferred() for x in range(10)]
        mockclient.send_fetch_request.side_effect = reqs_ds

        consumer = Consumer(mockclient, topic, part, mock_proc,
                            max_buffer_size=8 * FETCH_BUFFER_SIZE_BYTES)
        messages = [create_message(b'X' * (consumer.buffer_size * 9))]
        message_set = KafkaCodec._encode_message_set(messages, offset)
        d = consumer.start(offset)

        # Ok, we deliver only part of the message_set, up to the size requested
        while not d.called:
            # Get the buffer size from the last call
            request = mockclient.send_fetch_request.call_args[0][0][0]
            # Create a response only as large as the 'max_bytes' request param
            message_iter = KafkaCodec._decode_message_set_iter(
                message_set[0:request.max_bytes])
            responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 486,
                                       message_iter)]
            with patch.object(kconsumer, 'log'):
                consumer._request_d.callback(responses)
            # Advance the clock to trigger the next request
            clock.advance(0.01)

        self.failureResultOf(d, ConsumerFetchSizeTooSmall)
        consumer.stop()

    def test_consumer_fetch_response_with_wrong_partition(self):
        topic = 'fetch_response_with_wrong_partition'
        part = 68
        offset = 0
        mock_proc = Mock()
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        reqs_ds = [Deferred(), Deferred()]
        mockclient.send_fetch_request.side_effect = reqs_ds

        consumer = Consumer(mockclient, topic, part, mock_proc)
        d = consumer.start(offset)

        # Make sure the consumer started
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        self.assertEqual(consumer._request_d, reqs_ds[0])

        # create & deliver the responses
        messages = [create_message(b"v1", b"k1"), create_message(b"v2", b"k2")]
        message_set = KafkaCodec._encode_message_set(messages, offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        bad_messages = [create_message(b'fetch_response_with_wrong_partition')]
        bad_message_set = KafkaCodec._encode_message_set(bad_messages, offset)
        bad_message_iter = KafkaCodec._decode_message_set_iter(bad_message_set)
        responses = [
            FetchResponse(topic, part + 1, KAFKA_SUCCESS, 99, bad_message_iter),
            FetchResponse(topic, part,     KAFKA_SUCCESS, 99, message_iter),
        ]
        with patch.object(kconsumer, 'log') as klog:
            reqs_ds[0].callback(responses)
            klog.warning.assert_called_once_with(
                '%r: Got response with partition: %r not our own: %r',
                consumer, part + 1, part)
        # Make sure the processor was called
        self.assertTrue(mock_proc.called)

        self.assertEqual(1, consumer.stop())
        self.assertEqual(1, self.successResultOf(d))

    def test_consumer_do_fetch_not_reentrant(self):
        # This test is a bit of a hack to get coverage
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        consumer = Consumer(mockclient, 'do_fetch_not_reentrant', 8, Mock())
        d = consumer.start(0)
        request = FetchRequest('do_fetch_not_reentrant', 8, 0,
                               consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)

        # I think _do_fetch() cannot possibly (normally) be called when there's
        # an outstanding request, so force it
        with patch.object(kconsumer, 'log') as klog:
            consumer._do_fetch()
            klog.debug.assert_called_once_with(
                "_do_fetch: Outstanding request: %r", consumer._request_d)

        # And make sure no additional fetch request was made
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # clean up
        self.assertIsNone(consumer.stop())
        self.assertIsNone(self.successResultOf(d))

    def test_consumer_do_fetch_before_retry_call(self):
        # This test is a bit of a hack to get coverage
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_fetch_request.return_value = Deferred()
        consumer = Consumer(mockclient, 'do_fetch_before_retry_call', 8,
                            Mock())
        d = consumer.start(0)
        request = FetchRequest('do_fetch_before_retry_call', 8, 0,
                               consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # The error we'll return
        f = Failure(KafkaUnavailableError())  # Perhaps kafka wasn't up yet...

        # errback the request so the Consumer will create a _retry_call
        with patch.object(kconsumer, 'log'):
            consumer._request_d.errback(f)

        # I think _do_fetch() cannot possibly (normally) be called before the
        # retry_call fires, so force it
        with patch.object(kconsumer, 'log'):
            consumer._do_fetch()

        # clean up
        self.assertIsNone(consumer.stop())
        self.assertIsNone(self.successResultOf(d))

    def test_consumer_autocommit_during_commit(self):
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        commit_ds = [Deferred(), Deferred()]
        mockclient.send_offset_commit_request.side_effect = commit_ds
        the_group = 'XTC'
        the_topic = 'test_consumer_autocommit_during_commit'
        the_part = 119
        the_offset = 2496
        the_request = OffsetCommitRequest(
            the_topic, the_part, the_offset, TIMESTAMP_INVALID, None)
        # Create a consumer and muck with the state a bit...
        consumer = Consumer(mockclient, the_topic, the_part, Mock(), the_group)
        consumer._last_processed_offset = the_offset  # Fake processed msgs
        consumer._commit_looper = Mock()  # Mock a looping call to test reset
        d = consumer.commit()
        mockclient.send_offset_commit_request.assert_called_once_with(
            the_group, [the_request], consumer_id='', group_generation_id=-1)
        consumer._commit_looper.reset.assert_called_once_with()
        # Fake start_d, then force an auto-commit
        consumer._start_d = True
        consumer._auto_commit()
        # Check that still only one commit request has been made
        mockclient.send_offset_commit_request.assert_called_once_with(
            the_group, [the_request], consumer_id='', group_generation_id=-1)
        mockclient.send_offset_commit_request.reset_mock()
        # bump the last_processed_offset
        consumer._last_processed_offset = the_offset + 1  # Fake processed msgs
        # callback the first commit deferred.
        commit_response = [
            OffsetCommitResponse(the_topic, the_part, KAFKA_SUCCESS),
        ]
        commit_ds[0].callback(commit_response)
        self.assertTrue(d.called)
        # Check that the second commit request has been made
        the_request = OffsetCommitRequest(
            the_topic, the_part, the_offset + 1, TIMESTAMP_INVALID, None)
        mockclient.send_offset_commit_request.assert_called_once_with(
            the_group, [the_request], consumer_id='', group_generation_id=-1)

    def test_consumer_unhandled_commit_failure(self):
        """test_consumer_unhandled_commit_failure
        Test that if the commit() call's returned deferred encounters non-kafka
        errors that the commit attempt will not be retried
        """
        the_group = 'Bangles'
        the_topic = 'test_consumer_unhandled_commit_failure'
        the_part = 6
        the_offset = 4513
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        # Make the commit throw an error that won't allow request to be retried
        the_fail = Failure(ValueError(
            "test_consumer_unhandled_commit_failure induced failure"))
        commit_ds = [fail(the_fail), Deferred()]
        mockclient.send_offset_commit_request.side_effect = commit_ds

        # Create a consumer and muck with the state a bit...
        consumer = Consumer(mockclient, the_topic, the_part, Mock(),
                            the_group)
        consumer._last_processed_offset = the_offset  # Fake processed msgs

        # Patch the consumer's log so we can make sure the failure is logged
        with patch.object(kconsumer, 'log') as klog:
            commit_d = consumer.commit()

        # Make sure send_commit_request was called once, and error was logged
        the_request = OffsetCommitRequest(
            the_topic, the_part, the_offset, TIMESTAMP_INVALID, None)
        mockclient.send_offset_commit_request.assert_called_once_with(
            the_group, [the_request], consumer_id='', group_generation_id=-1)
        klog.error.assert_called_once_with(
            'Unhandleable failure during commit attempt: %r\n\t%r',
            ANY, ANY)
        self.assertEqual(self.failureResultOf(commit_d, ValueError), the_fail)
        # Eat the error
        commit_d.addErrback(lambda _: None)

    def test_consumer_commit_timer_failed(self):
        """test_consumer_commit_timer_failed
        Test that the looping call is restarted when an error occurs
        Somewhat artificial test to confirm that when failures occur in
        consumer._auto_commit (which cause the looping call to terminate) that
        the looping call is restarted.
        """
        the_group = 'Alphaville'
        the_topic = u'test_consumer_commit_timer_failed'
        the_part = 5
        the_offset = 5431
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        fetch_ds = [Deferred(), Deferred()]
        mockclient.send_fetch_request.side_effect = fetch_ds
        # Make the commit throw an error that won't allow request to be retried
        the_error = ValueError(
            "test_consumer_commit_timer_failed induced failure")

        # Create a consumer and muck with the state a bit...
        consumer = Consumer(mockclient, the_topic, the_part, Mock(),
                            the_group)
        consumer._last_processed_offset = the_offset  # Fake processed msgs

        # Start the consumer (starts auto-commit clock)
        start_d = consumer.start(the_offset)
        # Patch the consumer's log so we can make sure the failure is logged
        with patch.object(kconsumer, 'log') as klog:
            # Advance the clock to trigger auto-commit
            with patch.object(consumer, 'commit', side_effect=the_error):
                clock.advance(consumer.auto_commit_every_s)
        klog.warning.assert_called_once_with(
            '_commit_timer_failed: uncaught error %r: %s in _auto_commit',
            ANY, ANY)

        # Check that the looping call was restarted
        self.assertTrue(consumer._commit_looper.running)

        self.assertEqual(the_offset, consumer.stop())
        self.assertEqual(the_offset, self.successResultOf(start_d))

    def test_consumer_send_timer_stopped_error(self):
        # Purely for coverage
        client = Mock(reactor=MemoryReactorClock())
        consumer = Consumer(client, 'topic', 5, Mock(), 'abba')
        consumer.start(0)
        with patch.object(kconsumer, 'log') as klog:
            consumer._commit_timer_stopped('Borg')
        klog.warning.assert_called_once_with(
            '_commit_timer_stopped with wrong timer:%s not:%s', 'Borg',
            consumer._commit_looper)
        consumer.stop()

    def test_consumer_send_commit_request_not_concurrent(self):
        # Purely for coverage: Force a call of _send_commit_request
        # in order to effect the raise of OperationInProgress
        client = Mock()
        consumer = Consumer(client, 'topic', 5, Mock(), 'The Call')
        # Mess with the state
        consumer._last_processed_offset = 1
        the_mock = Mock()
        consumer._commit_req = the_mock
        self.assertRaises(OperationInProgress, consumer.commit)

    def test_consumer_shutdown_nothing_processing_no_cgroup(self):
        """
        Test the consumer shutdown happy path when no messages are currently
        being processed by the processor function (while waiting on fetch req),
        and further that there's no consumer group, so no commit needed
        """
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockproc = Mock()
        consumer = Consumer(mockclient, 'snpncgTopic', 1, mockproc)
        start_d = consumer.start(1)
        # Ensure a fetch request was made
        request = FetchRequest('snpncgTopic', 1, 1, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # Shutdown the consumer
        shutdown_d = consumer.shutdown()
        # Ensure the stop was signaled
        self.assertIsNone(self.successResultOf(start_d))
        # Ensure the shutdown was signaled
        self.assertIsNone(self.successResultOf(shutdown_d))
        # Ensure the processor was never called
        self.assertFalse(mockproc.called)

    def test_consumer_shutdown_nothing_processing(self):
        """
        Test the consumer shutdown happy path when no messages are currently
        being processed by the processor function (while waiting on fetch req).
        """
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockproc = Mock()
        consumer = Consumer(mockclient, 'snpTopic', 1, mockproc, 'snpGroup')
        start_d = consumer.start(1)
        # Ensure a fetch request was made
        request = FetchRequest('snpTopic', 1, 1, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # Shutdown the consumer
        shutdown_d = consumer.shutdown()
        # Ensure the stop was signaled
        self.assertIsNone(self.successResultOf(start_d))
        # Ensure the shutdown was signaled
        self.assertIsNone(self.successResultOf(shutdown_d))
        # Ensure the processor was never called
        self.assertFalse(mockproc.called)

    def test_consumer_shutdown_processing(self):
        """test_consumer_shutdown_processing
        Test the consumer shutdown happy path when messages are currently
        being processed by the processor function.
        """
        reqs_ds = [Deferred(), Deferred()]
        commit_d = [Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_fetch_request.side_effect = reqs_ds
        mockclient.send_offset_commit_request.side_effect = commit_d
        proc_d = Deferred()

        topic = 'tcsp'
        part = 2
        offset = 5

        consumer = Consumer(
            mockclient, topic, part, lambda *args, **kwargs: proc_d,
            'tcsp_group')
        start_d = consumer.start(offset)
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # create & deliver the response
        messages = [
            create_message(b"v1", b"k1"),
            create_message(b"v2", b"k2"),
        ]
        message_set = KafkaCodec._encode_message_set(messages, offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 99,
                                   message_iter)]
        reqs_ds[0].callback(responses)
        # Make sure the processor was called
        self.assertEqual(proc_d, consumer._processor_d)
        # While the consumer is waiting on the processor_d, shut it down
        shutdown_d = consumer.shutdown()
        # Ensure the processor deferred wasn't cancelled & consumer not stopped
        self.assertFalse(proc_d.called)
        self.assertNoResult(start_d)
        # complete the processor and ensure shutdown completed
        proc_d.callback(None)
        # Indicate a successful commit
        commit_d[0].callback(consumer._last_processed_offset)
        # Ensure the stop was signaled
        self.assertEqual(6, self.successResultOf(start_d))
        # Ensure the shutdown was signaled
        self.assertEqual(6, self.successResultOf(shutdown_d))
        self.assertEqual(6, consumer.last_processed_offset)
        self.assertEqual(6, consumer.last_committed_offset)

    def test_consumer_shutdown_commit_in_progress(self):
        """test_consumer_shutdown_commit_in_progress
        Test the consumer shutdown when there is a commit already in process
        In the case that consumer.shutdown() is called and there is already a
        shutdown request in flight to Kafka, ensure that the shutdown is tied
        to the successful commit of the last processed offset.
        """
        reqs_ds = [Deferred(), Deferred()]
        commit_ds = [Deferred(), Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_fetch_request.side_effect = reqs_ds
        mockclient.send_offset_commit_request.side_effect = commit_ds
        the_processor = Mock()
        proc_deferreds = [Deferred(), Deferred()]
        the_processor.side_effect = proc_deferreds

        topic = 'tcscip'
        part = 3
        offset = 6

        consumer = Consumer(
            mockclient, topic, part, the_processor, 'tcscip_group',
            auto_commit_every_n=2)
        start_d = consumer.start(offset)
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # create & deliver the response
        messages = [
            create_message(b"v1", b"k1"),
            create_message(b"v2", b"k2"),
        ]
        message_set = KafkaCodec._encode_message_set(messages, offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 99,
                                   message_iter)]
        reqs_ds[0].callback(responses)
        # Make sure the processor was called
        self.assertEqual(proc_deferreds[0], consumer._processor_d)
        # Ensure the processor deferred wasn't cancelled & consumer not stopped
        self.assertFalse(proc_deferreds[0].called)
        # complete the processor which will kick off auto-commit
        proc_deferreds[0].callback(None)
        # While the consumer is waiting on the commit reply, shut it down
        shutdown_d = consumer.shutdown()
        # Ensure consumer was not yet stopped/shutdown not complete
        self.assertFalse(start_d.called)
        self.assertFalse(shutdown_d.called)
        # Indicate a successful commit
        commit_ds[0].callback(consumer._last_processed_offset)
        # Ensure the stop was signaled
        self.assertEqual(7, self.successResultOf(start_d))
        # Ensure the shutdown was signaled
        self.assertEqual(7, self.successResultOf(shutdown_d))

    def test_consumer_shutdown_commit_failure(self):
        """test_consumer_shutdown_commit_failure
        Test the consumer shutdown when the commit attempt fails
        """
        reqs_ds = [Deferred(), Deferred()]
        commit_d = [Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_fetch_request.side_effect = reqs_ds
        mockclient.send_offset_commit_request.side_effect = commit_d
        proc_d = Deferred()

        topic = 'tcscf'
        part = 2
        offset = 5

        consumer = Consumer(
            mockclient, topic, part, lambda *args, **kwargs: proc_d,
            'tcscf_group')
        start_d = consumer.start(offset)
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # create & deliver the response
        messages = [
            create_message(b"v1", b"k1"),
            create_message(b"v2", b"k2"),
        ]
        message_set = KafkaCodec._encode_message_set(messages, offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 99,
                                   message_iter)]
        reqs_ds[0].callback(responses)
        # Make sure the processor was called
        self.assertEqual(proc_d, consumer._processor_d)
        # While the consumer is waiting on the processor_d, shut it down
        shutdown_d = consumer.shutdown()
        # Ensure the processor deferred wasn't cancelled & consumer not stopped
        self.assertNoResult(proc_d)
        # complete the processor and ensure shutdown completed
        proc_d.callback(None)
        # Indicate a failed commit
        the_fail = Failure(RuntimeError('Unretryable Commit Failure'))
        commit_d[0].errback(the_fail)
        # Ensure the stop was signaled with nothing committed
        self.assertEqual(6, self.successResultOf(start_d))
        self.assertEqual(6, consumer.last_processed_offset)
        self.assertIsNone(consumer.last_committed_offset)
        # Ensure the shutdown was signaled as an errback
        self.assertEqual(the_fail, self.failureResultOf(shutdown_d))

    def test_consumer_shutdown_processor_failure(self):
        """test_consumer_shutdown_processor_failure
        Test the consumer shutdown when the processor fails/errbacks after
        shutdown is called
        """
        reqs_ds = [Deferred(), Deferred()]
        commit_d = [Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_fetch_request.side_effect = reqs_ds
        mockclient.send_offset_commit_request.side_effect = commit_d
        the_processor = Mock()
        proc_deferreds = [Deferred(), Deferred()]
        the_processor.side_effect = proc_deferreds

        topic = 'tcspf'
        part = 3
        offset = 8

        consumer = Consumer(
            mockclient, topic, part, the_processor, 'tcspf_group',
            auto_commit_every_n=1)
        start_d = consumer.start(offset)
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # create & deliver the response
        messages = [
            create_message(b"v1", b"k1"),
            create_message(b"v2", b"k2"),
        ]
        message_set = KafkaCodec._encode_message_set(messages, offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 99,
                                   message_iter)]
        reqs_ds[0].callback(responses)
        # Make sure the processor was called
        self.assertEqual(proc_deferreds[0], consumer._processor_d)
        # While the consumer is waiting on the processor_d, shut it down
        shutdown_d = consumer.shutdown()
        # Ensure the processor deferred wasn't cancelled & consumer not stopped
        self.assertFalse(proc_deferreds[0].called)
        self.assertNoResult(start_d)
        # errback the processor
        the_fail = Failure(RuntimeError('Horrible Processor Failure'))
        proc_deferreds[0].errback(the_fail)
        # Ensure the stop was signaled with the failure
        self.assertEqual(self.failureResultOf(start_d), the_fail)
        # Ensure the shutdown was signaled as a callback, not errback
        self.assertIsNone(self.successResultOf(shutdown_d))
        self.assertIsNone(consumer.last_processed_offset)
        self.assertIsNone(consumer.last_committed_offset)

    def test_consumer_shutdown_processor_immediate_shutdown(self):
        """
        Test the consumer when the processor calls shutdown immediately.
        Any in-progress operations should be completed, and the start and
        shutdown deferreds should return the proper committed/processed offsets
        """
        reqs_ds = [Deferred(), Deferred()]
        commit_d = [Deferred()]
        proc_d_canceller = Mock()
        proc_d_errmock = Mock()
        proc_deferred = Deferred(canceller=proc_d_canceller)
        proc_deferred.addErrback(proc_d_errmock)
        proc_l = []  # Used to pass the shutdown_d out
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_fetch_request.side_effect = reqs_ds
        mockclient.send_offset_commit_request.side_effect = commit_d

        topic = 'tcspis'
        part = 5
        offset = 9

        def the_processor(consumer, messages):
            # Shutdown the consumer, and use proc_l to pass out the deferred
            d = consumer.shutdown()
            proc_l.append(d)
            # Even though we've called shutdown, return a deferred anyway...
            return proc_deferred

        consumer = Consumer(
            mockclient, topic, part, the_processor, 'tcspis_group',
            auto_commit_every_n=1)
        start_d = consumer.start(offset)
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # create & deliver the response
        messages = [
            create_message(b"v1", b"k1"),
            create_message(b"v2", b"k2"),
        ]
        message_set = KafkaCodec._encode_message_set(messages, offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 99,
                                   message_iter)]
        reqs_ds[0].callback(responses)
        # The processor shutdown the consumer prior to returning, so
        # stop/shutdown should have run, and the processor deferred
        # should have been cancelled. Check that's true
        proc_d_canceller.assert_called_once_with(proc_deferred)
        # Since the canceller didn't callback/errback, proc_deferred, it should
        # be errback'd by Twisted with a CancelledError
        commit_fail = proc_d_errmock.mock_calls[0][1][0]
        assert isinstance(commit_fail, Failure)
        commit_fail.trap(CancelledError)
        # Ensure the stop (start_d) was signaled with success
        self.assertIsNone(self.successResultOf(start_d))
        # Ensure the shutdown was signaled as a callback, not errback
        self.assertIsNone(self.successResultOf(proc_l[0]))

    def test_consumer_shutdown_called_twice(self):
        """
        Test the consumer shutdown when there is a shutdown already in progress
        """
        reqs_ds = [Deferred(), Deferred()]
        commit_d = [Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_fetch_request.side_effect = reqs_ds
        mockclient.send_offset_commit_request.side_effect = commit_d
        proc_d = Deferred()

        topic = 'csct'
        part = 2
        offset = 5

        consumer = Consumer(mockclient, topic, part, lambda *args, **kwargs: proc_d, 'csct_group')
        start_d = consumer.start(offset)
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # create & deliver the response
        messages = [
            create_message(b"v1", b"k1"),
            create_message(b"v2", b"k2"),
        ]
        message_set = KafkaCodec._encode_message_set(messages, offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        responses = [FetchResponse(topic, part, KAFKA_SUCCESS, 99,
                                   message_iter)]
        reqs_ds[0].callback(responses)
        # Make sure the processor was called
        self.assertEqual(proc_d, consumer._processor_d)
        # While the consumer is waiting on the processor_d, shut it down
        shutdown_d = consumer.shutdown()
        # Ensure the processor deferred wasn't cancelled & consumer not stopped
        self.assertFalse(proc_d.called)
        self.assertNoResult(start_d)
        self.assertFalse(shutdown_d.called)
        # While the consumer is waiting for the processor to complete, call
        # shutdown again and assert it raises a RestopError
        shutdown_d_2 = consumer.shutdown()
        the_fail = self.failureResultOf(shutdown_d_2, RestopError)
        self.assertEqual(the_fail.value.args, ("Shutdown called more than once.",))
        # Complete the shutdown.
        proc_d.callback(None)
        commit_d[0].callback(consumer._last_processed_offset)
        # Ensure the stop (start_d) was signaled with success
        self.assertEqual(6, self.successResultOf(start_d))
        # Ensure the shutdown was signaled as a callback, not errback
        self.assertEqual(6, self.successResultOf(shutdown_d))

    def test_consumer_shutdown_when_not_started(self):
        """
        Test the consumer shutdown when the consumer was never started
        """
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockproc = Mock()
        consumer = Consumer(mockclient, 'cswns', 1, mockproc)
        shutdown_d = consumer.shutdown()
        the_fail = self.failureResultOf(shutdown_d, RestopError)
        self.assertEqual(
            the_fail.value.args,
            ("Shutdown called on non-running consumer",))

    def test_consumer_commit_with_retrieved_offset(self):
        """
        Test that the consumer properly handles a commit operation from
        the processor function before deferred fires, after having retrieved
        committed offsets from Kafka. This tests that a bug which
        existed previously (<=v2.6.0) is fixed.
        """
        offset = 1234  # arbitrary, offset committed on topic
        fetch_offset = offset + 1  # fetch at next offset after committed
        highwatermark = offset + 100  # last message in topic/part
        topic = 'topic_with_committed_offsets'
        part = 56
        offset_fetch_ds = [Deferred()]
        fetch_ds = [Deferred(), Deferred()]
        proc_ds = [Deferred(), Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_offset_fetch_request.side_effect = offset_fetch_ds
        mockclient.send_fetch_request.side_effect = fetch_ds
        the_processor = Mock()
        the_processor.side_effect = proc_ds
        consumer = Consumer(mockclient, topic, part, the_processor,
                            consumer_group="myGroup",
                            auto_commit_every_n=0,
                            auto_commit_every_ms=0)
        start_d = consumer.start(OFFSET_COMMITTED)
        # Make sure request was made
        request = OffsetFetchRequest(topic, part)
        mockclient.send_offset_fetch_request.assert_called_once_with('myGroup', [request])
        # Deliver the response
        responses = [OffsetFetchResponse(topic, part, offset, b"METADATA",
                                         KAFKA_SUCCESS)]
        offset_fetch_ds[0].callback(responses)
        self.assertEqual(fetch_offset, consumer._fetch_offset)
        # Check that the message fetch was started
        request = FetchRequest(topic, part, fetch_offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # Fake the fetch response to trigger the processor call
        # create & deliver the response
        messages = [create_message(b"v1", b"k1")]
        message_set = KafkaCodec._encode_message_set(messages, fetch_offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)
        fetch_responses = [FetchResponse(topic, part, KAFKA_SUCCESS,
                                         highwatermark, message_iter)]
        fetch_ds[0].callback(fetch_responses)
        # Check that the processor function was properly called
        log.debug("MockClient Calls: %r ", mockclient.mock_calls)
        the_processor.assert_called_once_with(
            consumer,
            [SourcedMessage(topic, part, fetch_offset, messages[0])])
        # Attempt to commit offsets
        commit_d = consumer.commit()
        # the commit call should have short-circuited due to lack of
        # processing anything up to now.
        self.assertEqual(1234, self.successResultOf(commit_d))
        self.assertFalse(mockclient.send_offset_commit_request.called)
        # Stop the consumer to cleanup any outstanding operations
        self.assertIsNone(consumer.stop())
        self.assertIsNone(self.successResultOf(start_d))
        self.assertIsNone(consumer.last_processed_offset)
        self.assertEqual(1234, consumer.last_committed_offset)

    def test_consumer_consume_committed_no_offset_stored(self):
        """
        Test that when a consumer is started from OFFSET_COMMITTED and there
        is no committed offset that the fetch request is for OFFSET_EARLIEST,
        not 0 or any other offset

        https://github.com/ciena/afkak/issues/13
        """
        topic = u'notCommittedTopic'
        part = 0
        offset = 20170912
        group = u"aGroup"
        reqs_ds = [Deferred(), Deferred(), Deferred()]
        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_offset_fetch_request.return_value = reqs_ds[0]
        mockclient.send_offset_request.return_value = reqs_ds[1]
        mockclient.send_fetch_request.return_value = reqs_ds[2]
        consumer = Consumer(mockclient, topic, part, Mock(),
                            consumer_group=group)
        d = consumer.start(OFFSET_COMMITTED)
        # Make sure request for committed offset was made
        request = OffsetFetchRequest(topic, part)
        mockclient.send_offset_fetch_request.assert_called_once_with(group, [request])
        # Deliver the response. -1 offset, empty metadata
        responses = [OffsetFetchResponse(topic, part, -1, "", KAFKA_SUCCESS)]
        reqs_ds[0].callback(responses)
        self.assertEqual(OFFSET_EARLIEST, consumer._fetch_offset)
        # Make sure request for OFFSET_EARLIEST was made
        request = OffsetRequest(topic, part, OFFSET_EARLIEST, 1)
        mockclient.send_offset_request.assert_called_once_with([request])
        # Deliver the response. -1 offset, empty metadata
        responses = [OffsetResponse(topic, part, KAFKA_SUCCESS, [offset])]
        reqs_ds[1].callback(responses)
        self.assertEqual(offset, consumer._fetch_offset)
        # Check that the message fetch was started
        request = FetchRequest(topic, part, offset, consumer.buffer_size)
        mockclient.send_fetch_request.assert_called_once_with(
            [request], max_wait_time=consumer.fetch_max_wait_time,
            min_bytes=consumer.fetch_min_bytes)
        # Stop the consumer to cleanup any outstanding operations
        self.assertIsNone(consumer.stop())
        self.assertIsNone(self.successResultOf(d))

    def test_consumer_process_messages_should_exit_when_no_messages_left(self):
        client = Mock()
        a_partition = 9
        a_processor = Mock()
        a_consumer_group = 'My Consumer Group'

        consumer = Consumer(
            client, 'a_topic', a_partition, a_processor, a_consumer_group,
        )
        self.assertIsNone(self.successResultOf(consumer._process_messages([])))

    def test_consumer_process_messages_stack_safe(self):
        """
        The Consumer process a list of messages without causing a RecursionError
        Past versions of afkak used a callback chain to process fetched messages.
        """

        count = sys.getrecursionlimit() + 1

        class MessageProcessor(object):
            calls = 0

            def __call__(self, consumer, messages):
                self.calls += 1
                return succeed(None)

        # set up a Consumer and start it so it can consume messages

        offset = 0
        topic = 'some_topic'
        partition = 13
        consumer_group = "A nice day"

        message_list = [create_message(b'paylod', b'key') for _ in range(count)]
        message_set = KafkaCodec._encode_message_set(message_list, offset)
        message_iter = KafkaCodec._decode_message_set_iter(message_set)

        clock = MemoryReactorClock()
        mockclient = Mock(reactor=clock)
        mockclient.send_fetch_request.side_effect = [
            succeed([FetchResponse(topic, partition, KAFKA_SUCCESS, 5, message_iter)]),
        ]

        processor = MessageProcessor()

        # Instantiate a Consumer with auto_commit_every_n=1 to force batches of messages of
        # size 1 to be sent to the processor, in order to be able to make an assertion about the
        # the number of calls made to the message processor

        consumer = Consumer(
            mockclient, topic, partition, processor, consumer_group, auto_commit_every_n=1)
        consumer.start(offset)

        # A RecursionError during processing will cause the callback chain to stop, which
        # will result in messages without a corresponding call to the message processor

        self.assertEqual(count, processor.calls)

    def test_consumer_process_messages_should_notify_msg_block_when_no_messages_left(self):
        client = Mock()
        a_partition = 9
        a_processor = Mock()
        a_consumer_group = 'My Consumer Group'

        consumer = Consumer(
            client, 'a_topic', a_partition, a_processor, a_consumer_group,
        )
        d = consumer._msg_block_d = Deferred()
        consumer._process_messages([])
        self.assertTrue(self.successResultOf(d))



# -*- encoding: utf-8 -*-
# Copyright 2018 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os


def _twisted_debug():
    """
    When the ``AFKAK_TWISTED_DEBUG`` environment variable is set, enable
    debugging of deferreds and delayed calls.
    """
    if os.environ.get('AFKAK_TWISTED_DEBUG'):
        from twisted.internet import defer
        from twisted.internet.base import DelayedCall

        defer.setDebugging(True)
        DelayedCall.debug = True


def _nose_log_to_file():
    """
    Nose: still somehow a better test runner than Scalatest.
    """
    path = os.environ.get('AFKAK_TEST_LOG')
    if not path:
        return

    handler = logging.FileHandler(path)
    handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)8s %(threadName)8s'
                                           ' %(name)10s %(filename)s:%(lineno)d: %(message)s'))

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.addHandler(handler)


_nose_log_to_file()
_twisted_debug()



# -*- coding: utf-8 -*-
# Copyright 2014, 2015 Cyan, Inc.
# Copyright 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import os.path
import re
import shutil
import socket
import subprocess
import tempfile
import uuid
from datetime import datetime

from six.moves.urllib.parse import urlparse

from afkak.test.int.service import ExternalService, SpawnedService
from afkak.test.testutil import random_string


def get_open_port():
    sock = socket.socket()
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


class KafkaHarness(object):
    """
    `KafkaHarness` starts a Kafka cluster.

    :ivar str chroot:
        ZooKeeper chroot the Kafka brokers are configured to use.
    :ivar zk: `_ZookeeperFixture` instance
    :ivar brokers: `list` of `_KafakBroker` instances
    """
    @classmethod
    def start(cls, replicas, **kw):
        """
        Create and run a `KafkaHarness`.

        :param int replicas:
            Number of Kafka brokers to start. Must be an odd number.
        :param kw:
            Keyword arguments passed through to `_KafkaFixture`. The
            *replicas*, *zk_host*, *zk_port*, and *zk_chroot* are set
            automatically. *partitions* defaults to 8.
        """
        kw['zk_chroot'] = chroot = random_string(10)
        zk = _ZookeeperFixture.instance(chroot)
        kw['replicas'] = replicas
        kw.setdefault('partitions', 8)
        kw['zk_host'] = zk.host
        kw['zk_port'] = zk.port

        brokers = []
        for broker_id in range(replicas):
            brokers.append(_KafkaFixture.instance(broker_id, **kw))

        return cls(chroot, zk, brokers)

    def __init__(self, chroot, zk, brokers):
        """
        Private constructor.
        """
        self.chroot = chroot
        self.zk = zk
        self.brokers = brokers

    def halt(self):
        for broker in self.brokers:
            broker.close()
        self.zk.close()

    @property
    def bootstrap_hosts(self):
        """
        List of (host, port) tuples for the brokers.
        """
        return ["{}:{}".format(b.host, b.port) for b in self.brokers]


class _Fixture(object):
    kafka_version = os.environ.get('KAFKA_VERSION', '0.9.0.1')
    project_root = os.environ.get(
        'PROJECT_ROOT', os.path.abspath(
            os.path.join(os.path.dirname(__file__), "../..")))
    kafka_root = os.environ.get(
        "KAFKA_ROOT", os.path.join(project_root, 'servers',
                                   kafka_version, "kafka-bin"))

    @classmethod
    def test_resource(cls, filename):
        return os.path.join(cls.project_root, "servers", cls.kafka_version,
                            "resources", filename)

    @classmethod
    def kafka_run_class_args(cls, *args):
        result = [os.path.join(cls.kafka_root, 'bin', 'kafka-run-class.sh')]
        result.extend(args)
        return result

    @classmethod
    def kafka_run_class_env(cls):
        env = os.environ.copy()
        env['KAFKA_LOG4J_OPTS'] = "-Dlog4j.configuration=file:{}".format(
            cls.test_resource("log4j.properties"))
        return env

    @classmethod
    def render_template(cls, source_file, target_file, binding):
        with open(source_file, "r") as handle:
            template = handle.read()
        with open(target_file, "w") as handle:
            handle.write(template.format(**binding))


class _ZookeeperFixture(_Fixture):
    @classmethod
    def instance(cls, kafka_chroot):
        if "ZOOKEEPER_URI" in os.environ:  # pragma: no cover
            parse = urlparse(os.environ["ZOOKEEPER_URI"])
            (host, port) = (parse.hostname, parse.port)
            fixture = ExternalService(host, port)
        else:
            (host, port) = ("127.0.0.1", get_open_port())
            fixture = cls(host, port)

        fixture.open(kafka_chroot)
        return fixture

    def __init__(self, host, port):
        self.host = host
        self.port = port

        self.tmp_dir = None
        self.child = None
        self._log = logging.getLogger('fixtures.ZK')

    def __repr__(self):
        return "{}<{}:{}>".format(
            self.__class__.__name__,
            self.host,
            self.port,
        )

    def open(self, kafka_chroot):
        self.tmp_dir = tempfile.mkdtemp()
        properties_file = os.path.join(self.tmp_dir, "zookeeper.properties")

        properties = (
            "dataDir={tmp_dir}\n"
            "clientPort={port}\n"
            "clientPortAddress={host}\n"
            "maxClientCnxns=0\n"
            # Use reduced timeouts to speed up failovers.
            "tickTime=2000\n"
            "maxSessionTimeout=6000\n"
        ).format(
            tmp_dir=self.tmp_dir,
            host=self.host,
            port=self.port,
        )
        with open(properties_file, 'w') as f:
            f.write(properties)
        self._log.info("Running local instance with config:\n%s", properties)

        # Configure Zookeeper child process
        args = self.kafka_run_class_args("org.apache.zookeeper.server.quorum.QuorumPeerMain", properties_file)
        env = self.kafka_run_class_env()
        start_re = re.compile(
            "binding to port /{host}:|Starting server.*ZooKeeperServerMain"
            .format(host=re.escape(self.host)),
        )
        self._child = SpawnedService('zookeeper', self._log, args, env, start_re)
        self._child.start()

        self._log.debug("Creating ZooKeeper chroot node %r...", kafka_chroot)
        args = self.kafka_run_class_args(
            "org.apache.zookeeper.ZooKeeperMain",
            "-server", "%s:%d" % (self.host, self.port),
            "create",
            "/%s" % kafka_chroot,
            "afkak",
        )
        env = self.kafka_run_class_env()
        proc = subprocess.Popen(args, env=env, stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
        stdout, _stderr = proc.communicate()
        if proc.returncode != 0:  # pragma: no cover
            self._log.error("Failed to create Zookeeper chroot node. Process exited %d and output:\n%s",
                            proc.returncode, stdout.decode('utf-8', 'replace'))
            raise RuntimeError("Failed to create Zookeeper chroot node")
        self._log.debug("Done!")

    def close(self):
        self._log.debug("Stopping...")
        self._child.stop()
        self._child = None
        self._log.debug("Done!")
        shutil.rmtree(self.tmp_dir)


class _KafkaFixture(_Fixture):

    @classmethod
    def instance(cls, broker_id, zk_host, zk_port, zk_chroot, replicas,
                 partitions, message_max_bytes=1000000):
        if zk_chroot is None:
            zk_chroot = "afkak_" + str(uuid.uuid4()).replace("-", "_")
        if "KAFKA_URI" in os.environ:  # pragma: no cover
            parse = urlparse(os.environ["KAFKA_URI"])
            (host, port) = (parse.hostname, parse.port)
            fixture = ExternalService(host, port)
        else:
            (host, port) = ("127.0.0.1", get_open_port())
            fixture = cls(
                host=host, port=port, broker_id=broker_id, zk_host=zk_host,
                zk_port=zk_port, zk_chroot=zk_chroot, replicas=replicas,
                partitions=partitions, message_max_bytes=message_max_bytes,
            )
            fixture.open()
        return fixture

    def __init__(self, host, port, broker_id, zk_host, zk_port, zk_chroot,
                 replicas, partitions, message_max_bytes):
        self.host = host
        self.port = port
        self._log = logging.getLogger('fixtures.K{}'.format(broker_id))

        self.broker_id = broker_id

        self.zk_host = zk_host
        self.zk_port = zk_port
        self.zk_chroot = zk_chroot

        self.replicas = replicas
        self.partitions = partitions
        assert replicas % 2 == 1, "replica count must be odd, not {}".format(replicas)
        self.min_insync_replicas = replicas // 2 + 1

        self.message_max_bytes = message_max_bytes

        self.tmp_dir = None
        self.child = None
        self.running = False
        self.restartable = False  # Only restartable after stop() call

    def __repr__(self):
        return '{}<node_id={} {}:{} {}>'.format(
            self.__class__.__name__,
            self.broker_id,
            self.host,
            self.port,
            'running' if self.running else 'stopped',
        )

    def open(self):
        if self.running:  # pragma: no cover
            self._log.debug("Instance already running")
            return

        self.tmp_dir = tempfile.mkdtemp()
        self._properties_file = os.path.join(self.tmp_dir, "kafka.properties")
        self._log.info("Running local instance...")
        self._log.info("  host       = %s", self.host)
        self._log.info("  port       = %s", self.port)
        self._log.info("  broker_id  = %s", self.broker_id)
        self._log.info("  zk_host    = %s", self.zk_host)
        self._log.info("  zk_port    = %s", self.zk_port)
        self._log.info("  zk_chroot  = %s", self.zk_chroot)
        self._log.info("  replicas   = %s", self.replicas)
        self._log.info("  partitions = %s", self.partitions)
        self._log.info("  msg_max_sz = %s", self.message_max_bytes)
        self._log.info("  tmp_dir    = %s", self.tmp_dir)

        # Create directories
        os.mkdir(os.path.join(self.tmp_dir, "logs"))
        os.mkdir(os.path.join(self.tmp_dir, "data"))

        # Generate configs
        template = self.test_resource("kafka.properties")
        self.render_template(template, self._properties_file, vars(self))

        self._log.debug("Starting...")
        self.child = self._make_child()
        self.child.start()
        self._log.debug("Done!")
        self.running = True

    def _make_child(self):
        """
        Configure Kafka child process
        """
        name = 'kafka{}'.format(self.broker_id)
        args = self.kafka_run_class_args("kafka.Kafka", self._properties_file)
        env = self.kafka_run_class_env()
        # Match a message like (0.9.0.1 and earlier):
        #
        #     [2018-07-17 18:06:00,915] INFO [Kafka Server 0], started (kafka.server.KafkaServer)
        #
        # Or like this (1.1.1):
        #
        #     [2018-09-27 17:23:46,818] INFO [KafkaServer id=0] started (kafka.server.KafkaServer)
        start_re = re.compile((
            r"("
            r"\[Kafka Server {broker_id}\], [Ss]tarted"
            r"|"
            r"\[KafkaServer id={broker_id}\] started"
            r")"
        ).format(broker_id=self.broker_id))
        return SpawnedService(name, self._log, args, env, start_re)

    def close(self):
        if not self.running:  # pragma: no cover
            self._log.warning("Instance already stopped")
            return

        self._log.info("Stopping... %s at %s",
                       self.tmp_dir, datetime.utcnow().isoformat())
        self.child.stop()
        self.child = None
        self._log.info("Done!")
        shutil.rmtree(self.tmp_dir)
        self.running = False

    def stop(self):
        """Stop/cleanup the child SpawnedService"""
        self.child.stop()
        self.child = None
        self._log.info("Child stopped.")
        self.running = False
        self.restartable = True

    def restart(self):
        """Start a new child SpawnedService with same settings and tmpdir"""
        if not self.restartable:  # pragma: no cover
            self._log.error("*** Kafka [%s:%d]: Restart attempted when not stopped.",
                            self.host, self.port)
            return
        self.child = self._make_child()
        self._log.debug("Starting...")
        self.child.start()
        self._log.debug("Done!")
        self.running = True
        self.restartable = False



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2017, 2018, 2019 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from itertools import cycle
from random import randint

try:
    from pyhash import murmur2_32 as _murmur2_32
    _c_murmur2 = _murmur2_32(0x9747b28c)
except ImportError:  # pragma: no cover
    warnings.warn(
        "Import of pyhash failed, using pure python", ImportWarning,
    )
    _c_murmur2 = None


def pure_murmur2(byte_array, seed=0x9747b28c):
    """Pure-python Murmur2 implementation.

    Based on java client, see org.apache.kafka.common.utils.Utils.murmur2
    https://github.com/apache/kafka/blob/0.8.2/clients/src/main/java/org/apache/kafka/common/utils/Utils.java#L244
    Args:
        byte_array: bytearray - Raises TypeError otherwise

    Returns: MurmurHash2 of byte_array bytearray
    Raises: TypeError if byte_array arg is not of type bytearray
    """
    # Ensure byte_array arg is a bytearray
    if not isinstance(byte_array, bytearray):
        raise TypeError("Type: %r of 'byte_array' arg must be 'bytearray'",
                        type(byte_array))

    length = len(byte_array)
    # 'm' and 'r' are mixing constants generated offline.
    # They're not really 'magic', they just happen to work well.
    m = 0x5bd1e995
    r = 24
    mod32bits = 0xffffffff

    # Initialize the hash to a random value
    h = seed ^ length
    length4 = length // 4

    for i in range(length4):
        i4 = i * 4
        k = ((byte_array[i4 + 0] & 0xff) + ((byte_array[i4 + 1] & 0xff) << 8) +
             ((byte_array[i4 + 2] & 0xff) << 16) + ((byte_array[i4 + 3] & 0xff) << 24))
        k &= mod32bits
        k *= m
        k &= mod32bits
        k ^= (k % 0x100000000) >> r  # k ^= k >>> r
        k &= mod32bits
        k *= m
        k &= mod32bits

        h *= m
        h &= mod32bits
        h ^= k
        h &= mod32bits

    # Handle the last few bytes of the input array
    extra_bytes = length % 4
    if extra_bytes == 3:
        h ^= (byte_array[(length & ~3) + 2] & 0xff) << 16
        h &= mod32bits

    if extra_bytes >= 2:
        h ^= (byte_array[(length & ~3) + 1] & 0xff) << 8
        h &= mod32bits

    if extra_bytes >= 1:
        h ^= (byte_array[length & ~3] & 0xff)
        h &= mod32bits
        h *= m
        h &= mod32bits

    h ^= (h % 0x100000000) >> 13  # h >>> 13;
    h &= mod32bits
    h *= m
    h &= mod32bits
    h ^= (h % 0x100000000) >> 15  # h >>> 15;
    h &= mod32bits

    return h


class Partitioner(object):
    """
    Base class for a partitioner

    :ivar bytes topic: Topic name
    :ivar partitions: :class:`list` of :class:`int`
    """
    def __init__(self, topic, partitions):
        """
        Initialize the partitioner

        partitions - A list of available partitions (during startup)
        """
        self.topic = topic
        self.partitions = partitions

    def partition(self, key, partitions):
        """
        Takes a key (string) and partitions (list) as argument and returns
        a partition to be used for the message

        partitions - The list of partitions is passed in every call. This
                     may look like an overhead, but it will be useful
                     (in future) when we handle cases like rebalancing
        """
        raise NotImplementedError('partition function has to be implemented')


class RoundRobinPartitioner(Partitioner):
    """
    Implements a round robin partitioner which sends data to partitions
    in a round robin fashion. Also supports starting each new partitioner
    at a random offset into the cycle of partitions
    """
    randomStart = False

    @classmethod
    def set_random_start(cls, randomStart):
        cls.randomStart = randomStart

    def __init__(self, topic, partitions):
        super(RoundRobinPartitioner, self).__init__(topic, partitions)
        self._set_partitions(partitions)

    def __repr__(self):
        return '<RoundRobinPartitioner {}:{}>'.format(self.randomStart,
                                                      self.partitions)

    def _set_partitions(self, partitions):
        self.partitions = sorted(partitions)
        self.iterpart = cycle(partitions)
        if self.randomStart:
            for _ in range(randint(0, len(partitions) - 1)):
                next(self.iterpart)

    def partition(self, key, partitions):
        # Refresh the partition list if necessary
        if self.partitions != partitions:
            self._set_partitions(partitions)
        return next(self.iterpart)


class HashedPartitioner(Partitioner):
    """
    Implements a partitioner which selects the target partition based on
    the hash of the key.
    """
    if _c_murmur2:
        def _hash(self, key):
            """
            Coerce the input into `bytes` with UTF-8 encoding for the
            C MurmurHash2 implementation.
            """
            if isinstance(key, type(u'')):
                key = key.encode('UTF-8')
            elif isinstance(key, bytearray):
                key = bytes(key)
            elif not isinstance(key, bytes):
                raise TypeError('Partition key {!r} must be {} or {},'
                                ' not {}'.format(key, type(b''), type(u''), type(key)))
            return _c_murmur2(key)
    else:
        def _hash(self, key):
            """
            Coerce the input into a bytearray for the pure-Python MurmurHash2
            implementation.
            """
            if isinstance(key, type(u'')):
                key = bytearray(key, 'UTF-8')
            elif isinstance(key, bytes):
                key = bytearray(key)
            elif not isinstance(key, bytearray):
                raise TypeError('Partition key {!r} must be str, bytes, or bytearray, not {}'.format(key, type(key)))
            return pure_murmur2(key)

    def partition(self, key, partitions):
        """
        Select a partition based on the hash of the key.

        :param key: Partition key
        :type key: text string or UTF-8 `bytes` or `bytearray`
        :param list partitions:
            An indexed sequence of partition identifiers.
        :returns:
            One of the given partition identifiers. The result will be the same
            each time the same key and partition list is passed.
        """
        return partitions[(self._hash(key) & 0x7FFFFFFF) % len(partitions)]



# -*- coding: utf-8 -*-
# Copyright (C) 2015 Cyan, Inc.

import unittest


class TestPackage(unittest.TestCase):
    def test_top_level_namespace(self):
        import afkak as afkak1
        self.assertEqual(afkak1.KafkaClient.__name__, "KafkaClient")
        self.assertEqual(afkak1.client.__name__, "afkak.client")
        self.assertEqual(afkak1.codec.__name__, "afkak.codec")

    def test_submodule_namespace(self):
        import afkak.client as client1
        self.assertEqual(client1.__name__, "afkak.client")
        self.assertEqual(client1.KafkaClient.__name__, "KafkaClient")

        from afkak import client as client2
        self.assertEqual(client2.__name__, "afkak.client")
        self.assertEqual(client2.KafkaClient.__name__, "KafkaClient")

        from afkak.client import KafkaClient as KafkaClient1
        self.assertEqual(KafkaClient1.__name__, "KafkaClient")

        from afkak.codec import gzip_encode as gzip_encode1
        self.assertEqual(gzip_encode1.__name__, "gzip_encode")

        from afkak import KafkaClient as KafkaClient2
        self.assertEqual(KafkaClient2.__name__, "KafkaClient")

        from afkak.codec import snappy_encode
        self.assertEqual(snappy_encode.__name__, "snappy_encode")



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2018 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging

from twisted.internet.defer import Deferred, fail
from twisted.internet.protocol import Factory, connectionDone
from twisted.logger import Logger
from twisted.protocols.basic import Int32StringReceiver

"""
Twisted protocols that understand Kafka message framing
"""

log = logging.getLogger('afkak.protocol')
log.addHandler(logging.NullHandler())


class _BaseKafkaProtocol(Int32StringReceiver):
    MAX_LENGTH = 2 ** 31 - 1  # Max a signed Int32 can represent


class KafkaProtocol(_BaseKafkaProtocol):
    """
    Very thin wrapper around the Int32StringReceiver
    Simply knows to call its factory.handleResponse()
    method with the string received by stringReceived() and
    to cleanup the factory reference when the connection is lost
    """
    factory = None

    def stringReceived(self, string):
        self.factory.handleResponse(string)

    def connectionLost(self, reason=connectionDone):
        self.factory._connectionLost(reason)
        self.factory = None

    def lengthLimitExceeded(self, length):
        log.error("Broker at %s sent a %d byte message, exceeding the size limit of %d. "
                  "Terminating connection.", self.transport.getPeer(), length,
                  self.MAX_LENGTH)
        self.transport.loseConnection()


class KafkaBootstrapProtocol(_BaseKafkaProtocol):
    """
    `KafkaBootstrapProtocol` sends and receives Kafka messages.

    It knows just enough about Kafka message framing to correlate responses
    with requests. A deferred is issued for every request and fires when
    a response is received or the connection is lost.

    :ivar dict _pending:
        Map of correlation ID to Deferred.
    """
    _log = Logger()

    def connectionMade(self):
        self._pending = {}
        self._failed = None

    def stringReceived(self, response):
        """
        Handle a response from the broker.
        """
        correlation_id = response[0:4]
        try:
            d = self._pending.pop(correlation_id)
        except KeyError:
            self._log.warn((
                "Response has unknown correlation ID {correlation_id!r}."
                " Dropping connection to {peer}."
            ), correlation_id=correlation_id, peer=self.transport.getPeer())
            self.transport.loseConnection()
        else:
            d.callback(response)

    def connectionLost(self, reason=connectionDone):
        """
        Mark the protocol as failed and fail all pending operations.
        """
        self._failed = reason
        pending, self._pending = self._pending, None
        for d in pending.values():
            d.errback(reason)

    def lengthLimitExceeded(self, length):
        self._log.error(
            "Broker at {peer} sent a {length:,d} byte message, exceeding the size limit of {max_length:,d}.",
            peer=self.transport.getPeer(), length=length, max_length=self.MAX_LENGTH,
        )
        self.transport.loseConnection()

    def request(self, request):
        """
        Send a request to the Kafka broker.

        :param bytes request:
            The bytes of a Kafka `RequestMessage`_ structure. It must have
            a unique (to this connection) correlation ID.

        :returns:
            `Deferred` which will:

              - Succeed with the bytes of a Kafka `ResponseMessage`_
              - Fail when the connection terminates

        .. _RequestMessage:: https://kafka.apache.org/protocol.html#protocol_messages

        """
        if self._failed is not None:
            return fail(self._failed)
        correlation_id = request[4:8]
        assert correlation_id not in self._pending
        d = Deferred()
        self.sendString(request)
        self._pending[correlation_id] = d
        return d


bootstrapFactory = Factory.forProtocol(KafkaBootstrapProtocol)



# -*- coding: utf-8 -*-
# Copyright 2015 Cyan, Inc.
# Copyright 2018 Ciena Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import unittest

import mock
from twisted.internet.address import IPv4Address
from twisted.internet.error import ConnectionDone, ConnectionLost
from twisted.logger import LogLevel, globalLogPublisher
from twisted.python.failure import Failure
from twisted.test.iosim import FakeTransport
from twisted.test.proto_helpers import (
    EventLoggingObserver, StringTransportWithDisconnection,
)
from twisted.trial.unittest import SynchronousTestCase

from .._protocol import KafkaBootstrapProtocol, KafkaProtocol
from .logtools import capture_logging


class TheFactory(object):
    """
    `TheFactory` implements the bits of `_KafkaBrokerClient` that
    `_KafkaProtocol` interacts with.
    """
    log = logging.getLogger(__name__).getChild('TheFactory')

    def handleResponse(self, string):
        """Called for each response."""

    def _connectionLost(self, reason):
        """Called when the connection goes down."""


class KafkaProtocolTests(unittest.TestCase):
    """Test `afkak._protocol.KafkaProtocol`"""

    def test_stringReceived(self):
        """
        The factory is notified of message receipt.
        """
        kp = KafkaProtocol()
        kp.factory = factory_spy = mock.Mock(wraps=TheFactory())

        kp.stringReceived(b"testing")

        factory_spy.handleResponse.assert_called_once_with(b"testing")

    def test_connectionLost_cleanly(self):
        """
        The factory is notified of connection loss.
        """
        kp = KafkaProtocol()
        kp.factory = factory_spy = mock.Mock(wraps=TheFactory())
        reason = Failure(ConnectionDone())

        kp.connectionLost(reason)

        factory_spy._connectionLost.assert_called_once_with(reason)
        self.assertIsNone(kp.factory)

    def test_lengthLimitExceeded(self):
        """
        An error is logged and the connection dropped when an oversized message
        is received.
        """
        too_long = KafkaProtocol.MAX_LENGTH + 1
        peer = IPv4Address('TCP', '1.2.3.4', 1234)
        kp = KafkaProtocol()
        kp.factory = factory_spy = mock.Mock(wraps=TheFactory())
        kp.transport = StringTransportWithDisconnection(peerAddress=peer)
        kp.transport.protocol = kp

        with capture_logging(logging.getLogger('afkak.protocol')) as records:
            kp.lengthLimitExceeded(too_long)

        self.assertEqual(1, len(factory_spy._connectionLost.mock_calls))

        [record] = records
        record.getMessage()  # Formats okay.
        self.assertEqual((
            'Broker at %s sent a %d byte message, exceeding the size limit of %d. '
            'Terminating connection.'
        ), record.msg)
        self.assertEqual((peer, too_long, kp.MAX_LENGTH), record.args)


class KafkaBootstrapProtocolTests(SynchronousTestCase):
    """Test `afkak._protocol.KafkaBootstrapProtocol`

    :ivar peer: Peer IAddress, an IPv4 address
    :ivar protocol: `KafkaBootstrapProtocol` object, connected in setUp
    :ivar transport: `FakeTransport` object associated with the protocol,
        connected in setUp
    """
    def setUp(self):
        self.peer = IPv4Address('TCP', 'kafka', 9072)
        self.protocol = KafkaBootstrapProtocol()
        self.transport = FakeTransport(self.protocol, isServer=False, peerAddress=self.peer)
        self.protocol.makeConnection(self.transport)

    def test_one_request(self):
        """
        Happy path: a request is made and a response received.
        """
        correlation_id = b'corr'
        client_request = b'api ' + correlation_id + b'corr more stuff'

        d = self.protocol.request(client_request)
        self.assertNoResult(d)

        # The request was written to the server.
        server_request = self.transport.getOutBuffer()
        self.assertEqual(b'\0\0\0\x17' + client_request, server_request)

        self.transport.bufferReceived(b'\0\0\0\x05' + correlation_id + b'y')
        self.assertEqual(correlation_id + b'y', self.successResultOf(d))

    def test_disconnected(self):
        """
        Pending and future requests fail when the connection goes away.
        """
        d = self.protocol.request(b'api corr stuff')
        self.assertNoResult(d)

        self.transport.disconnectReason = ConnectionLost('Bye.')
        self.transport.reportDisconnect()
        self.failureResultOf(d, ConnectionLost)

        self.failureResultOf(self.protocol.request(b'api corr more'), ConnectionLost)

    def test_unknown_correlation_id(self):
        """
        A warning is logged and the connection dropped when a response with an
        unknown correlation ID is received.
        """
        events = EventLoggingObserver.createWithCleanup(self, globalLogPublisher)
        self.transport.bufferReceived(b'\0\0\0\x101234 more stuff..')
        self.assertTrue(self.transport.disconnecting)

        [event] = events
        self.assertEqual(LogLevel.warn, event['log_level'])
        self.assertEqual(self.peer, event['peer'])
        self.assertEqual(b'1234', event['correlation_id'])

    def test_oversized_response(self):
        """
        An oversized response from the server prompts disconnection.
        """
        d = self.protocol.request(b'api corr blah blah')

        self.transport.bufferReceived(b'\xff\xff\xff\xff')  # 2**32 - 1, way too large.

        self.assertTrue(self.transport.disconnecting)
        self.assertNoResult(d)  # Will fail when the disconnect completes.
